import json
import time
import threading
import functools
import itertools
import traceback
import collections

import vqt.hotkeys as vq_hotkey
import vqt.saveable as vq_save
import envi.expression as e_expr
import envi.qt.memory as e_mem_qt
import envi.memcanvas as e_memcanvas
import envi.qt.memory as e_qt_memory
import vivisect.qt.views as viv_q_views
import envi.qt.memcanvas as e_qt_memcanvas

import visgraph.layouts.dynadag as vg_dynadag

import vivisect.base as viv_base
import vivisect.renderers as viv_rend
import vivisect.qt.memory as vq_memory
import vivisect.qt.ctxmenu as vq_ctxmenu
import vivisect.tools.graphutil as viv_graphutil

from PyQt5 import Qt, QtCore, QtGui, QtWebEngine, QtWidgets
from PyQt5.QtCore import pyqtSignal
from PyQt5.QtWidgets import *

from envi.threads import firethread
from vqt.main import idlethread, eatevents, workthread, vqtevent

from vqt.common import *
from vivisect.const import *

class VQVivFuncgraphCanvas(vq_memory.VivCanvasBase):
    paintUp = pyqtSignal()
    paintDown = pyqtSignal()
    paintMerge = pyqtSignal()
    refreshSignal = pyqtSignal()

    def __init__(self, *args, **kwargs):
        vq_memory.VivCanvasBase.__init__(self, *args, **kwargs)
        self.curs = QtGui.QCursor()

    # These have changed because QtWebEngine suxxs: https://bugreports.qt.io/browse/QTBUG-43602
    def event(self, evt):
        if evt.type() == Qt.QEvent.ChildAdded:
            evt.child().installEventFilter(self)
        elif evt.type() == Qt.QEvent.ChildRemoved:
            evt.child().removeEventFilter(self)
        return vq_memory.VivCanvasBase.event(self, evt)

    def eventFilter(self, src, evt):
        if evt.type() == Qt.QEvent.Wheel:
            return self._wheelEvent(evt)
        if evt.type() == Qt.QEvent.MouseMove:
            # Intercept mouse movements because frickin qt broke those for our shift scrolling,
            # but return False so we don't block the event from propagating to other event handlers
            # (this was the cause of the edge lines not highlighting on mouse over)
            self._mouseMoveEvent(evt)
            return False
        return vq_memory.VivCanvasBase.eventFilter(self, src, evt)

    def _wheelEvent(self, event):
        mods = QApplication.keyboardModifiers()
        if mods == QtCore.Qt.ShiftModifier:
            delta = event.angleDelta().y()
            factord = delta / 1000.0
            self.setZoomFactor(self.zoomFactor() + factord)
            event.accept()
            return True

        # e_qt_memcanvas.VQMemoryCanvas.wheelEvent(self, event)
        return False

    def _setMousePos(self, data):
        self.curs.setPos(*self.basepos)

    def _mouseMoveEvent(self, event):
        mods = QApplication.keyboardModifiers()
        if mods == QtCore.Qt.ShiftModifier:
            x = event.globalX()
            y = event.globalY()
            if self.lastpos:
                dx = -(x - self.lastpos[0])
                dy = -(y - self.lastpos[1])
                # dx = x - self.lastpos[0]
                # dy = y - self.lastpos[1]
                # TODO: Just use scrollPosition()
                self.page().runJavaScript(f'window.scrollBy({dx}, {dy});', self._setMousePos)
            else:
                self.lastpos = (x, y)
                self.basepos = (x, y)

            event.accept()
            return
        self.lastpos = None
        return e_qt_memcanvas.VQMemoryCanvas.mouseMoveEvent(self, event)

    def _renderMemoryFinish(self, cb, data):
        cb(data)

    def _renderMemoryCallback(self, cb, data):
        if not data:
            return

        for va, size in data[:-1]:
            va = int(va)
            size = int(size)
            sel = '#codeblock_%.8x' % va
            # DEV: this cannot be partialmethod. It *has* to be callable
            #runner = functools.partial(self._renderMemoryFinish, cb)
            e_memcanvas.MemoryCanvas.renderMemory(self, va, size, cb=None, clear=False, sel=sel)

        va, size = data[-1]
        va = int(va)
        size = int(size)
        sel = '#codeblock_%.8x' % va
        e_memcanvas.MemoryCanvas.renderMemory(self, va, size, cb=cb, clear=False, sel=sel)

    def renderMemory(self, blocks, cb=None):
        '''
        Funcgraph specific renderMemory() function.
        '''
        if not cb:
            cb = self.__nopcb

        js = '''var nodes = %s
        canv = document.querySelector("#memcanvas");
        if (canv != null) {
            var addend = ''
            for (const [va, size] of nodes) {
                var selector = 'codeblock_' + va.toString(16).padStart(8, '0')
                var node = document.getElementById("${selector}");
                if (node == null) {
                    var elem = document.createElement('div')
                    elem.id = selector
                    elem.className = 'codeblock'
                    canv.appendChild(elem)
                } else {
                    node.innerHTML = ''
                }
            }
        }
        nodes
        ''' % (blocks,)
        runner = functools.partial(self._renderMemoryCallback, cb)
        self.page().runJavaScript(js, runner)

    def __nopcb(self):
        pass

    def contextMenuEvent(self, event):
        tag = self._canv_curtag
        if self._canv_curva is not None:
            menu = vq_ctxmenu.buildContextMenu(self.vw, va=self._canv_curva, parent=self, nav=self.parent(), tag=tag)
        else:
            menu = QMenu(parent=self)

        self.viewmenu = menu.addMenu('view   ')
        self.viewmenu.addAction("Save frame to HTML", ACT(self._menuSaveToHtml))
        self.viewmenu.addAction("Refresh", ACT(self.refresh))
        self.viewmenu.addAction("Paint Up", ACT(self.paintUp.emit))
        self.viewmenu.addAction("Paint Down", ACT(self.paintDown.emit))
        self.viewmenu.addAction("Paint Down until remerge", ACT(self.paintMerge.emit))

        menu.exec_(event.globalPos())

    def _navExpression(self, expr):
        if self._canv_navcallback:
            self._canv_navcallback(expr)

    def refresh(self):
        '''
        Redraw the function graph (actually, tells the View to do it)
        '''
        self.reload()

    @idlethread
    def setScrollPosition(self, x, y):
        '''
        Sets the view reticle to an absolute scroll position
        '''
        self.page().runJavaScript(f'window.scroll({x}, {y})')
        eatevents()

def reset():
    VQVivFuncgraphView.viewidx = itertools.count()

class VQVivFuncgraphView(vq_hotkey.HotKeyMixin, e_qt_memory.EnviNavMixin, QWidget, vq_save.SaveableWidget, viv_base.VivEventCore):
    _renderDoneSignal = pyqtSignal()

    viewidx = itertools.count()

    def __init__(self, vw, vwqgui):
        self.vw = vw
        self.fva = None
        self.graph = None
        self.nodes = []
        self._xref_cache = set()
        self._xref_cache_bg = set()
        self._xref_cache_lock = threading.Lock()

        self.vwqgui = vwqgui
        self._last_viewpt = None
        self._rendering = False
        self._renderWaiting = False
        self._renderlock = threading.Lock()
        self.history = collections.deque((), 100)

        self._autorefresh = None # created in getRendToolsMenu()

        self._leading = False
        self._following = None
        self._follow_menu = None  # init'd in handler below

        QWidget.__init__(self, parent=vwqgui)
        vq_hotkey.HotKeyMixin.__init__(self)
        viv_base.VivEventCore.__init__(self, vw)
        e_qt_memory.EnviNavMixin.__init__(self)
        self.setEnviNavName('FuncGraph%d' % next(self.viewidx))

        self._renderDoneSignal.connect(self._refresh_cb)

        self.top_box = QWidget(parent=self)
        hbox = QHBoxLayout(self.top_box)
        hbox.setContentsMargins(2, 2, 2, 2)
        hbox.setSpacing(4)

        self.histmenu = QMenu(parent=self)
        self.histmenu.aboutToShow.connect(self._histSetupMenu)

        self.hist_button = QPushButton('History', parent=self.top_box)
        self.hist_button.setMenu(self.histmenu)

        self.addr_entry = QLineEdit(parent=self.top_box)

        self.rend_tools = QPushButton('Opts', parent=self.top_box)
        self.rend_tools.setMenu( self.getRendToolsMenu() )

        self.mem_canvas = VQVivFuncgraphCanvas(vw, syms=vw, parent=self)
        self.mem_canvas.setNavCallback(self.enviNavGoto)
        self.mem_canvas.refreshSignal.connect(self.refresh)
        self.mem_canvas.paintUp.connect(self._hotkey_paintUp)
        self.mem_canvas.paintDown.connect(self._hotkey_paintDown)
        self.mem_canvas.paintMerge.connect(self._hotkey_paintMerge)

        self.loadDefaultRenderers()

        self.addr_entry.returnPressed.connect(self._nav_expr)

        hbox.addWidget(self.hist_button)
        hbox.addWidget(self.addr_entry)
        hbox.addWidget(self.rend_tools)

        vbox = QVBoxLayout(self)
        vbox.setContentsMargins(4, 4, 4, 4)
        vbox.setSpacing(4)
        vbox.addWidget(self.top_box)
        vbox.addWidget(self.mem_canvas, stretch=100)

        self.top_box.setLayout(hbox)

        self.setLayout(vbox)
        self.updateWindowTitle()

        # Do these last so we are all setup...
        vwqgui.addEventCore(self)
        vwqgui.vivMemColorSignal.connect(self.mem_canvas._applyColorMap)

        QtWidgets.QShortcut(QtGui.QKeySequence("Escape"), self, activated=self._hotkey_histback, context=3)

        # TODO: Transition these to the above pattern (since escape/ctrl-c)
        # See: https://stackoverflow.com/questions/56890831/qwidget-cannot-catch-escape-backspace-or-c-x-key-press-events
        self.addHotKey('ctrl+0', 'funcgraph:resetzoom')
        self.addHotKeyTarget('funcgraph:resetzoom', self._hotkey_resetzoom)
        self.addHotKey('ctrl+=', 'funcgraph:inczoom')
        self.addHotKeyTarget('funcgraph:inczoom', self._hotkey_inczoom)
        self.addHotKey('ctrl+-', 'funcgraph:deczoom')
        self.addHotKeyTarget('funcgraph:deczoom', self._hotkey_deczoom)
        self.addHotKey('f5', 'funcgraph:refresh')
        self.addHotKeyTarget('funcgraph:refresh', self.refresh)
        self.addHotKey('ctrl+U', 'funcgraph:paintup')
        self.addHotKeyTarget('funcgraph:paintup', self._hotkey_paintUp)
        self.addHotKey('ctrl+D', 'funcgraph:paintdown')
        self.addHotKeyTarget('funcgraph:paintdown', self._hotkey_paintDown)
        self.addHotKey('ctrl+M', 'funcgraph:paintmerge')
        self.addHotKeyTarget('funcgraph:paintmerge', self._hotkey_paintMerge)
        self.addHotKey('x', 'viv:xrefsto')
        self.addHotKeyTarget('viv:xrefsto', self._viv_xrefsto)

    def toggleAutoRefresh(self):
        '''
        This allows the function graph to auto-refresh on changes (or not)
        '''
        self._autorefresh.setChecked(not self._autorefresh.isChecked())

    def _canHazAutoRefresh(self):
        '''
        Should we autorefresh on updates?
        '''
        return self._autorefresh.isChecked()

    def rendToolsSetName(self, user=None):
        curname = self.getEnviNavName()
        mwname, ok = QInputDialog.getText(self, 'Set Mem Window Name', 'Name', text=curname)
        if ok:
            self.setMemWindowName(str(mwname))

        if self.vw.server and self._leading:
            if user is None:
                user = self.vw.config.user.name
            self.vw.modifyLeaderSession(self.uuid, user, mwname)
        
    def rendToolsMenu(self, event):
        menu = self.getRendToolsMenu()
        menu.exec_(self.mapToGlobal(self.rend_tools.pos()))

    def getRendToolsMenu(self):
        menu = QMenu(parent=self.rend_tools)
        menu.addAction('set name', self.rendToolsSetName)

        self._autorefresh = QAction('auto-refresh', menu, checkable=True, checked=True)
        menu.addAction(self._autorefresh)

        if self.vw.server:
            leadact = QAction('lead', menu, checkable=True)

            def leadToggle():
                self._leading = not self._leading
                # We can only follow if not leading... (deep huh? ;) )
                self._follow_menu.setEnabled(not self._leading)
                if self._leading:
                    self._following = None
                    locexpr = self.addr_entry.text()
                    self.vw.iAmLeader(self.uuid, self.getEnviNavName(), locexpr)
                else:
                    self.vw.killLeaderSession(self.uuid)
                self.updateWindowTitle()

            def clearFollow():
                self._following = None
                self.updateWindowTitle()

            leadact.toggled.connect(leadToggle)
            menu.addAction(leadact)
            self._follow_menu = menu.addMenu('Follow..')

            self._rtmRebuild()

        return menu

    def _rtmAddLeaderSession(self, uuid, user, fname):
        '''
        Add Action to RendToolsMenu (Opts/Follow) for a given session
        '''
        logger.info("_rtmAddLeaderSession(%r, %r, %r)", uuid, user, fname)

        def setFollow():
            self._following = uuid
            self.updateWindowTitle()
            self.navToLeader()

        action = self._follow_menu.addAction('%s - %s' % (user, fname), setFollow)
        action.setWhatsThis(uuid)

    def _rtmModLeaderSession(self, uuid, user, fname):
        '''
        Add Action to RendToolsMenu (Opts/Follow) for a given session
        '''
        logger.info("_rtmModLeaderSession(%r, %r, %r)", uuid, user, fname)

        action = self._rtmGetActionByUUID(uuid)
        action.setText('%s - %s' % (user, fname))
        if self._following == uuid:
            self.updateWindowTitle()

    def _rtmGetActionByUUID(self, uuid):
        for action in self._follow_menu.actions():
            if action.whatsThis() == uuid:
                return action

        logger.info("Unknown UUID: %r" % uuid)

    def _rtmDelLeaderSession(self, uuid):
        '''
        Remove Action item from RendToolsMenu (Opts/Follow)
        '''
        action = self._rtmGetActionByUUID(uuid)
        if action:
            self._follow_menu.removeAction(action)
            logger.info("Removing %r from Follow menu (%r)" % (action, uuid))
            if self._following == uuid:
                self._following = None
                self.updateWindowTitle()
        else:
            logger.warning("Attempting to remove Menu Action that doesn't exist: %r", uuid)
            self._rtmRebuild()

    def _rtmRebuild(self):
        '''
        Sync current menu system with vw.leaders
        '''
        self._follow_menu.clear()

        def clearFollow():
            self._following = None
            self.updateWindowTitle()

        self._follow_menu.addAction('(disable)', clearFollow)

        # add in the already existing sessions...
        for uuid, (user, fname) in self.vw.getLeaderSessions().items():
            self._rtmAddLeaderSession(uuid, user, fname)

    def _rtmGetUUIDs(self):
        '''
        Returns a list of active UUIDs for leader sessions
        '''
        return [action.whatsThis() for action in self._follow_menu.actions()]

    def _nav_expr(self):
        expr = self.addr_entry.text()
        self.history.append(expr)
        self._renderMemory()

    def _hotkey_histback(self):
        if len(self.history) >= 2:
            self.history.pop()
            expr = self.history.pop()
            self.enviNavGoto(expr)

    def _hotkey_resetzoom(self):
        self.mem_canvas.setZoomFactor(1)

    def _hotkey_inczoom(self):
        newzoom = self.mem_canvas.zoomFactor()
        if 1 > newzoom > .75:
            newzoom = 1
        elif newzoom < .5:
            newzoom += .125
        else:
            newzoom += .25

        if newzoom < 0:
            return

        self.mem_canvas.setZoomFactor(newzoom)

    def _hotkey_deczoom(self):
        newzoom = self.mem_canvas.zoomFactor()
        if newzoom <= .5:
            newzoom -= .125
        else:
            newzoom -= .25

        self.mem_canvas.setZoomFactor(newzoom)

    def _viv_xrefsto(self):

        if self.mem_canvas._canv_curva is not None:
            xrefs = self.vw.getXrefsTo(self.mem_canvas._canv_curva)
            if len(xrefs) == 0:
                self.vw.vprint('No xrefs found!')
                return

            title = 'Xrefs To: 0x%.8x' % self.mem_canvas._canv_curva
            view = viv_q_views.VQXrefView(self.vw, self.vwqgui, xrefs=xrefs, title=title)
            dock = self.vwqgui.vqDockWidget(view, floating=True)
            dock.resize(800, 600)

    @firethread
    def refresh(self):
        '''
        Cause the Function Graph to redraw itself.
        This is particularly helpful because comments and name changes don't
        immediately display.  Perhaps someday this will update only the blocks
        that have changed since last update, and be fast, so we can update
        after every change.
        '''
        # debounce logic
        if self._renderWaiting:
            # if we are already rendering... and another thing is waiting
            #logger.debug('(%r) someone is already waiting... returning', threading.current_thread())
            return

        while self._rendering:
            #logger.debug('(%r) waiting for render to complete...', threading.current_thread())
            self._renderWaiting = True
            time.sleep(.1)

        self._renderWaiting = False

        with self._renderlock:
            #logger.debug('(%r) render beginning...', threading.current_thread())
            self._rendering = True

        # actually do the refresh
        self._last_viewpt = self.mem_canvas.page().scrollPosition()
        # TODO: history should track this as well and return to the same place
        self.clearText()
        self.fva = None
        self._renderMemory()

    @workthread
    def _refresh_cb(self):
        '''
        This is a hack to make sure that when _renderMemory() completes,
        _refresh_cb() gets run after all other rendering events yet to come.
        '''
        if self._last_viewpt is None:
            return

        self.mem_canvas.setScrollPosition(self._last_viewpt.x(), self._last_viewpt.y())
        self._last_viewpt = None

        with self._renderlock:
            self._rendering = False
        #logger.debug('(%r) render finished!', threading.current_thread())

        # update the "real" _xref_cache from the one we just built
        with self._xref_cache_lock:
            self._xref_cache = self._xref_cache_bg.copy()

    def _histSetupMenu(self):
        self.histmenu.clear()

        history = []
        for expr in self.history:
            try:
                addr = self.vw.parseExpression(expr)
                menustr = '0x%.8x' % addr
                sym = self.vw.getSymByAddr(addr)
                if sym is not None:
                    menustr += ' - %s' % repr(sym)
            except e_expr.ExpressionFail:
                menustr = "UNKNOWN-Exception (%r)" % expr

            history.append((menustr, expr))

        history.reverse()
        for menustr,expr in history:
            self.histmenu.addAction(menustr, ACT(self._histSelected, expr))

    def _histSelected(self, expr):
        while self.history.pop() != expr:
            pass
        self.enviNavGoto(expr)

    def isReferenced(self, va):
        '''
        During function building, we can grab any xrefs from this function and cache them
        This is the easy way to determine if an address is referenced from this function
        '''
        with self._xref_cache_lock:
            if va in self._xref_cache:
                return True

        # catch any "stragglers"
        if va in self._xref_cache_bg:
            return True

        return False

    def _refreshIfNecessary(self, va):
        '''
        Called from VWE event handlers with updated VA
        '''
        if not self._canHazAutoRefresh():
            return

        if self.vw.getFunction(va) == self.fva:
            self.refresh()

        if self.isReferenced(va):
            self.refresh()

    def VWE_SYMHINT(self, vw, event, einfo):
        va, idx, hint = einfo
        self.mem_canvas.renderMemoryUpdate(va, 1)
        self._refreshIfNecessary(va)

    def VWE_ADDLOCATION(self, vw, event, einfo):
        va, size, ltype, tinfo = einfo
        self.mem_canvas.renderMemoryUpdate(va, size)
        self._refreshIfNecessary(va)

    def VWE_DELLOCATION(self, vw, event, einfo):
        va, size, ltype, tinfo = einfo
        self.mem_canvas.renderMemoryUpdate(va, size)
        self._refreshIfNecessary(va)

    def VWE_ADDFUNCTION(self, vw, event, einfo):
        va, meta = einfo
        self.mem_canvas.renderMemoryUpdate(va, 1)
        self._refreshIfNecessary(va)

    def VWE_SETFUNCMETA(self, vw, event, einfo):
        fva, key, val = einfo
        self._refreshIfNecessary(fva)

    def VWE_SETFUNCARGS(self, vw, event, einfo):
        fva, fargs = einfo
        self._refreshIfNecessary(fva)

    def VWE_COMMENT(self, vw, event, einfo):
        va, cmnt = einfo
        self.mem_canvas.renderMemoryUpdate(va, 1)
        self._refreshIfNecessary(va)

    @idlethread
    def VWE_SETNAME(self, vw, event, einfo):
        va, name = einfo
        self.mem_canvas.renderMemoryUpdate(va, 1)
        for fromva, tova, rtype, rflag in self.vw.getXrefsTo(va):
            self.mem_canvas.renderMemoryUpdate(fromva, 1)
        self._refreshIfNecessary(va)

    @idlethread
    def VTE_IAMLEADER(self, vw, event, einfo):
        # we have to maintain the list here, since the menu itself is not 
        # autogenerated each time, like a context menu.
        uuid, user, fname, locexpr = einfo
        self._rtmAddLeaderSession(uuid, user, fname)

    @idlethread
    def VTE_KILLLEADER(self, vw, event, einfo):
        uuid = einfo
        self._rtmDelLeaderSession(uuid)

    @idlethread
    def VTE_FOLLOWME(self, vw, event, einfo):
        uuid, expr = einfo
        if self._following != uuid:
            return
        self.enviNavGoto(expr)

    @idlethread
    def VTE_MODLEADER(self, vw, event, einfo):
        uuid, user, fname = einfo
        self._rtmModLeaderSession(uuid, user, fname)

    @idlethread
    def enviNavGoto(self, expr, sizeexpr=None):
        if self._leading:
            self.vw.followTheLeader(self.uuid, str(expr))

        self.addr_entry.setText(expr)
        self.history.append( expr )
        self.updateWindowTitle()
        self._renderMemory()

    @idlethread
    def navToLeader(self):
        uuid = self._following
        logger.debug("navToLeader(%r)", uuid)
        if uuid:
            expr = self.vw.getLeaderLoc(uuid)
            if not expr:
                return

            self.enviNavGoto(expr)

    def vqGetSaveState(self):
        return { 'expr':str(self.addr_entry.text()), 'hist':json.dumps(list(self.history))}

    def vqSetSaveState(self, state):
        hist = state.get('hist')
        if hist:
            self.history = collections.deque(json.loads(hist), maxlen=100)

        expr = state.get('expr','')
        self.enviNavGoto(expr)

    def setMemWindowName(self, mwname):
        '''
        Set the memory window name/title prefix to the given string.
        '''
        self.setEnviNavName(mwname)
        self.updateWindowTitle()

    def getExprTitle(self):
        va = -1
        expr = str(self.addr_entry.text())

        try:

            va = self.vw.parseExpression(expr)
            smartname = self.vw.getName(va, smart=True)
            title = '%s (0x%x)' % (smartname, va)

            name = self.vw.getName(va)
            if name is not None:
                title = name

        except Exception:
            title = 'expr error'

        if self._leading:
            title += ' (leading)'

        if self._following is not None:
            uuid = self._following
            user, window = self.vw.getLeaderInfo(uuid)
            title += ' (following %s %s)' % (user, window)

        return title, va

    def updateWindowTitle(self, data=None):
        ename = self.getEnviNavName()
        expr, va = self.getExprTitle()
        self.setWindowTitle('%s: %s' % (ename, expr))
        return va

    # DEV: None of these methods are meant to be called directly by anybody but themselves,
    # since they're setup in a way to make renderFunctionGraph play nicely with pyqt5
    def _finishFuncRender(self, data):
        '''
        Update the window title and emit the renderDoneSignal so other things can run that
        are sitting on that signal
        '''
        addr = self.updateWindowTitle()
        if addr is not None:
            vqtevent('viv:colormap', {addr: 'orange'})

        self._renderDoneSignal.emit()

    def _edgesDone(self, data):
        '''
        Almost done. All this does is scroll the selected virtual address into view.
        '''
        addr = self.updateWindowTitle()
        if addr is None:
            return
        self.mem_canvas.page().runJavaScript('''
        var node = document.getElementsByName("viv:0x%.8x")[0];
        if (node != null) {
            node.scrollIntoView();
        }
        ''' % addr, self._finishFuncRender)

    def _layoutEdges(self, data):
        '''
        So...before this, we would only highlight part of the edge lines. That's because
        The dynadag layout class shoves a bunch of ghost nodes and edges into the graph
        class member, and each of those end up getting their own html polyline element.

        It'd be a massive change to the Dynadag class to rip those out, and those edges are necessary.
        So instead, we take care of things here. What the edges/nodes end up looking like in
        the graph after the dynadag layout is:
        (VA1, VA2)
        (VA3, GHOST_NODE1)
        (GHOST_NODE1, GHOST_NODE2)
        (GHOST_NODE2, GHOST_NODE3)
        (GHOST_NODE3, VA3)

        Where a bunch of ghost edges for layout purposes were inserted between the codeblocks for
        VA3 and VA4. So this function deals with that via creating a bunch of paths, keeping
        in mind the possibility that an edge line can possibly split for things like switch
        statements, and links all the points on the edge paths together to make one big
        polyline.

        This is done in two stages. First loop is to identify the possible series of ghost edges,
        and only graph the starting point, where we have (Actual_VA1, Ghost_Node). The other
        cases, like (Ghost_Node, Actual_VA2), (Ghost_Node, Ghost_Node) are ignored, since the
        second loop picks those up, and the (Actual_VA1, Actual_VA2) is dealt with since that's the
        base, super easy just do the thing case.

        The second loop just follows all the edge lines and ghost edges to make the one big polyline
        for each edge so the entire thing can be highlighted via the plineover function we've got
        stashed in envi/qt/html.py
        '''
        edgejs = ''
        svgid = 'funcgraph_%.8x' % self.fva
        graph = self.graph
        todo = []
        for eid, n1, n2, einfo in graph.getEdges():
            src = graph.getNode(n1)
            dst = graph.getNode(n2)
            points = einfo.get('edge_points')

            # if neither are ghosts, cool. just make the edge
            if not src[1].get('ghost', False) and not dst[1].get('ghost', False):
                pointstr = ' '.join(['%d,%d' % (x, y) for (x, y) in points])
                edgejs += f'drawSvgLine("{svgid}", "edge_%.8s", "{pointstr}");' % eid
                continue

            # if both are ghosts, w/e. skip it. we'll pick it up later
            if src[1].get('ghost', False) and dst[1].get('ghost', False):
                continue

            # if n1 is a ghost and n2 is not, that's fine. We'll pick it up later
            if src[1].get('ghost', False) and not dst[1].get('ghost', False):
                continue

            # ok. juicy case
            # if n1 is not a ghost and n2 is, we gotta build all the possible paths from this guy out
            todo.append((eid, n1, n2, einfo))

        # Build out the pointstr lines starting from the concrete n1, following the ghost n2's
        for edge in todo:
            eid, n1, n2, einfo = edge
            points = einfo.get('edge_points')
            splits = [([n1, n2], points)]

            while splits:
                path, points = splits.pop()
                node = graph.getNode(path[-1])

                # we've hit the end of this chain, finish the points chain
                # and add the completed points to a done list
                if not node[1].get('ghost', False):
                    pointstr = ' '.join(['%d,%d' % (x, y) for (x, y) in points])
                    edgejs += f'drawSvgLine("{svgid}", "edge_%.8s", "{pointstr}");' % eid
                    continue

                # Otherwise, deal with splits
                for ref in graph.getRefsFrom(node):
                    eid, n1, n2, einfo = ref
                    newpoints = einfo.get('edge_points')
                    path.append(n2)
                    points.extend(newpoints)
                    splits.append((path, points))

        self.mem_canvas.page().runJavaScript(edgejs, self._edgesDone)

    def _layoutDynadag(self, data):
        '''
        This actually lays codeblocks out on the memory canvas where they should be
        '''
        for nid, nprops in self.graph.getNodes():
            width, height = data[str(nid)]
            self.graph.setNodeProp((nid, nprops), "size", (width, height+7))
        self.dylayout = vg_dynadag.DynadagLayout(self.graph, barry=20)
        self.dylayout.layoutGraph()

        width, height = self.dylayout.getLayoutSize()

        svgid = 'funcgraph_%.8x' % self.fva
        svgjs = f'svgwoot("vbody", "{svgid}", {width+18}, {height});'
        for nid, nprops in self.graph.getNodes():
            cbva = nprops.get('cbva')
            if cbva is None:
                continue

            xpos, ypos = nprops.get('position')

            foid = 'fo_cb_%.8x' % cbva
            cbid = 'codeblock_%.8x' % cbva

            js = f'''
            var node = document.getElementById("{cbid}");
            addSvgForeignObject("{svgid}", "{foid}", node.offsetWidth+16, node.offsetHeight+7);
            addSvgForeignHtmlElement("{foid}", "{cbid}");
            moveSvgElement("{foid}", {xpos}, {ypos});
            '''
            svgjs += js

        self.mem_canvas.page().runJavaScript(svgjs, self._layoutEdges)

    def _getNodeSizes(self, data):
        '''
        Actually grab all the sizes of the codeblocks that we renderd in the many calls to
        _renderCodeBlock. runJavaScript has some limited ability to return values from
        javascript land to python town, so in this case, we're shoving the offsetWidth
        and offsetHeight of each of the codeblocks into a dictionary that _layoutDynadag
        can reach into to get the sizes so it can set them for use in the line layout stuff

        As with a couple other of these functions, the data param exists to make PyQt5's function
        call ing happy. We don't do anything with it.
        '''
        js = '''
        var sizes = {};
        '''

        for nid, nprops in self.graph.getNodes():
            try:
                cbname = 'codeblock_%.8x' % nid
            except Exception as e:
                self.vw.vprint('Failed to build cbname during funcgraph building: %s' % str(e))
                return
            js += f'''
            sizes[{nid}] = [document.getElementById("{cbname}").offsetWidth, document.getElementById("{cbname}").offsetHeight];
            '''
        js += 'sizes;'
        self.mem_canvas.page().runJavaScript(js, self._layoutDynadag)

    def _renderCodeBlock(self, data):
        '''
        Render a codeblock to the canvas. self.mem_canvas.renderMemory ends up having
        to run a bunch of javascript to render the block to the screen. So we do them
        one at a time, chaining callbacks together with some state in self.nodes to let us
        know when we've rendered all the codeblocks in the function.

        One day we'll optimize this to be one big blob of JS. But not today. But this could use
        some safety rails if the user switches functions in the middle of rendering

        Side note: the data param is unused. It exists to shut PyQt5 up.
        '''
        if len(self.nodes):
            # render codeblock
            blocks = []
            for node in self.nodes:
                cbva = node[1].get('cbva')
                cbsize = node[1].get('cbsize')
                blocks.append([cbva, cbsize])  # yes this has to be a list so javascript plays nicely

                # update _xref_cache_bg (made "real" when rendering complete)
                endva = cbva + cbsize
                while cbva < endva:
                    for xref in self.vw.getXrefsFrom(cbva):
                        xrfr, xrto, xrtype, xrflag = xref
                        self._xref_cache_bg.add(xrto)

                    lva, lsz, ltp, ltinfo = self.vw.getLocation(cbva)
                    cbva += lsz
            self.mem_canvas.renderMemory(blocks, self._getNodeSizes)

        else:
            self._getNodeSizes(None)

    def renderFunctionGraph(self, fva=None, graph=None):
        '''
        Begins the process of drawing the function graph to the canvas.

        So, this is a bit of complicated mess due to how PyQt5's runJavaScript method works.
        runJavaScript is asynchronous, but not like actual python async, but Qt's async variant
        with their event loop. The only way to get some level of a guarantee that the
        javascript ran (which we need for getting the size of codeblocks and line layouts) is
        via the secondary parameter, which is a callback that gets run when the javascript
        completes. That being said, technically according to their docs, the callback is guaranteed
        to be run, but it might be during page destruction. In practice it's somewhat responsive.
        runJavascript also doesn't check if the DOM has been created. So...yea. In practice
        that doesn't matter too much, but something to keep in mind.

        So the general method is to build up a bunch of javascript that we need to run in order
        to render the codeblocks to get their sizes, layout the graph lines, realign everything
        nicely, etc. And it's all callbacks, all the way down.
        '''
        if fva is not None:
            self.fva = fva

        if graph is None:
            try:
                graph = viv_graphutil.buildFunctionGraph(self.vw, fva, revloop=True)
            except Exception as e:
                self.vw.vprint(f'Error building function graph for {fva} ({str(e)}')
                self.vw.vprint(traceback.format_exc())
                return

        self.graph = graph

        # Clear _xref_cache
        self._xref_cache_bg.clear()

        # Go through each of the nodes and render them so we know sizes (and xrefs)
        self.nodes = self.graph.getNodes()
        self._renderCodeBlock(None)

    def _renderedSameFva(self, data):
        addr = self.updateWindowTitle()
        if addr is not None:
            vqtevent('viv:colormap', {addr: 'orange'})

    @idlethread
    def _renderMemory(self):
        try:

            expr = str(self.addr_entry.text())
            if not expr:
                return

            try:
                addr = self.vw.parseExpression(expr)
            except Exception as e:
                self.mem_canvas.addText('Invalid Address: %s (%s)' % (expr, e))
                return

            # get a location anchor if one exists, otherwise, we may end up in no-man's land,
            # since we rely on labels, which only exist for the base of a location.
            loc = self.vw.getLocation(addr)
            if loc is not None:
                addr = loc[L_VA]

            # check if we're already rendering this function. if so, just scroll to addr
            fva = self.vw.getFunction(addr)
            if fva is None:
                self.vw.vprint('0x%.8x is not in a function!' % addr)
                return

            page = self.mem_canvas.page()
            if fva == self.fva:
                page.runJavaScript('''
                document.getElementsByName("viv:0x%.8x")[0].scrollIntoView();
                %d;
                ''' % (addr, addr), self._renderedSameFva)
                return
            # if we're rendering a different function, get to work!
            self.clearText()
            self.renderFunctionGraph(fva)

        except Exception as e:
            self.vw.vprint('_renderMemory hit exception: %s' % str(e))
            self.vw.vprint('%s' % traceback.format_exc())

    def loadDefaultRenderers(self):
        vivrend = viv_rend.WorkspaceRenderer(self.vw)
        self.mem_canvas.addRenderer('Viv', vivrend)
        self.mem_canvas.setRenderer('Viv')

    def clearText(self):
        # Pop the svg and reset #memcanvas
        js = ''
        if self.fva is not None:
            svgid = 'funcgraph_%.8x' % self.fva
            js += '''
            var node = document.getElementById("%s");
            if (node != null) {
                node.remove();
            }
            ''' % svgid

        js += '''
        var canv = document.querySelector("#memcanvas");
        if (canv != null) {
            canv.innerHTML = "";
        }
        '''
        self.mem_canvas.page().runJavaScript(js)

    def _hotkey_paintUp(self, va=None):
        '''
        Paint the VA's from the selected basic block up to all possible
        non-looping starting points.
        '''
        graph = viv_graphutil.buildFunctionGraph(self.vw, self.fva, revloop=True)
        startva = self.mem_canvas._canv_curva
        if startva is None:
            return

        viv_graphutil.preRouteGraphUp(graph, startva, mark='hit')

        count = 0
        colormap = {}
        for node in graph.getNodesByProp('hit'):
            count += 1
            off = 0
            cbsize = node[1].get('cbsize')
            if cbsize is None:
                raise Exception('node has no cbsize: %s' % repr(node))

            # step through opcode for a node
            while off < cbsize:
                op = self.vw.parseOpcode(node[0] + off)
                colormap[op.va] = 'orange'
                off += len(op)

        self.vw.vprint("Colored Blocks: %d" % count)
        vqtevent('viv:colormap', colormap)
        return colormap

    def _hotkey_paintDown(self, va=None):
        '''
        Paint the VA's from the selected basic block down to all possible
        non-looping blocks.  This is valuable for determining what code can
        execute from any starting basic block, without a loop.
        '''
        # TODO: make overlapping colors available for multiple paintings

        graph = viv_graphutil.buildFunctionGraph(self.vw, self.fva, revloop=True)
        startva = self.mem_canvas._canv_curva
        if startva is None:
            return

        viv_graphutil.preRouteGraphDown(graph, startva, mark='hit')

        count = 0
        colormap = {}
        for node in graph.getNodesByProp('hit'):
            count += 1
            off = 0
            cbsize = node[1].get('cbsize')
            if cbsize is None:
                raise Exception('node has no cbsize: %s' % repr(node))

            # step through opcode for a node
            while off < cbsize:
                op = self.vw.parseOpcode(node[0] + off)
                colormap[op.va] = 'brown'
                off += len(op)

        self.vw.vprint("Colored Blocks: %d" % count)
        vqtevent('viv:colormap', colormap)
        return colormap

    def _hotkey_paintMerge(self, va=None):
        '''
        same as paintdown but only until the graph remerges
        '''

        graph = viv_graphutil.buildFunctionGraph(self.vw, self.fva, revloop=True)
        startva = self.mem_canvas._canv_curva
        if startva is None:
            return

        viv_graphutil.findRemergeDown(graph, startva)

        count = 0
        colormap = {}
        for node in graph.getNodesByProp('hit'):
            count += 1
            off = 0
            cbsize = node[1].get('cbsize')
            if cbsize is None:
                raise Exception('node has no cbsize: %s' % repr(node))

            # step through opcode for a node
            while off < cbsize:
                op = self.vw.parseOpcode(node[0] + off)
                colormap[op.va] = 'brown'
                off += len(op)

        self.vw.vprint("Colored Blocks: %d" % count)
        vqtevent('viv:colormap', colormap)
        return colormap




import vstruct
from vstruct.primitives import *
import vstruct.primitives as vs_prim

class fat_header(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.magic = v_uint32(bigend=True)
        self.nfat_arch = v_uint32(bigend=True)

class fat_arch(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.cputype    = v_uint32(bigend=True)  # cpu specifier (int) */
        self.cpusubtype = v_uint32(bigend=True)  # machine specifier (int) */
        self.offset     = v_uint32(bigend=True)  # file offset to this object file */
        self.size       = v_uint32(bigend=True)  # size of this object file */
        self.align      = v_uint32(bigend=True)  # alignment as a power of 2 */

class fat_arch_64(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.cputype    = v_uint32(bigend=True)  # cpu specifier (int) */
        self.cpusubtype = v_uint32(bigend=True)  # machine specifier (int) */
        self.offset     = v_uint64(bigend=True)  # file offset to this object file */
        self.size       = v_uint64(bigend=True)  # size of this object file */
        self.align      = v_uint32(bigend=True)  # alignment as a power of 2 */
        self.reserved   = v_uint32(bigend=True)



'''
A package for a basic command interpreter for the graph.
(and graph db...)
'''

import sys
import cmd

import visgraph.graphcore as vg_gcore


class GraphCli(cmd.Cmd):

    def __init__(self, graph=None):
        cmd.Cmd.__init__(self)
        if graph is None:
            graph = vg_gcore.Graph()
        self.graph = graph

    def do_addnode(self, line):
        '''
        Add a node with the given key=value properties on the CLI.
        (the property nid is special and MUST be an integer)
        '''
        return self.graph.addNode()

    def do_quit(self, line):
        '''
        Exit the visgraph cli....
        '''
        raise SystemExit()


def main():
    cli = GraphCli()
    cli.cmdloop()


if __name__ == '__main__':
    sys.exit(main())



'''
Home for the registered emulators of different types...
'''
import vivisect.impemu.platarch.h8 as v_i_h8
import vivisect.impemu.platarch.arm as v_i_arm
import vivisect.impemu.platarch.a64 as v_i_a64
import vivisect.impemu.platarch.i386 as v_i_i386
import vivisect.impemu.platarch.amd64 as v_i_amd64
import vivisect.impemu.platarch.msp430 as v_i_msp430

import vivisect.impemu.platarch.linux as v_i_linux
import vivisect.impemu.platarch.windows as v_i_windows

workspace_emus  = {
    'h8': v_i_h8.H8WorkspaceEmulator,
    'a64' :v_i_a64.A64WorkspaceEmulator,
    'arm': v_i_arm.ArmWorkspaceEmulator,
    'i386': v_i_i386.i386WorkspaceEmulator,
    'amd64': v_i_amd64.Amd64WorkspaceEmulator,
    'msp430': v_i_msp430.Msp430WorkspaceEmulator,
    'thumb': v_i_arm.ThumbWorkspaceEmulator,
    'thumb16': v_i_arm.Thumb16WorkspaceEmulator,
    ('windows', 'i386'): v_i_windows.Windowsi386Emulator,
    ('windows', 'amd64'): v_i_windows.WindowsAmd64Emulator,
    ('linux', 'i386'): v_i_linux.Linuxi386Emulator,
    ('linux', 'amd64'): v_i_linux.LinuxAmd64Emulator,
}



import platform
# This module needs to be "relocatable"

def getHostId():
    osname = platform.system().lower()
    if osname == 'darwin':
        import darwinhostid
        return darwinhostid.getHostId()

    elif osname in ['microsoft','windows']:
        import windowshostid
        return windowshostid.getHostId()

    return None




import sys
import logging
import functools

from queue import Queue
from threading import current_thread

from PyQt5 import QtCore
from PyQt5.QtWidgets import *

import envi.threads as e_threads

logger = logging.getLogger(__name__)

def idlethread(func):
    '''
    A decorator which causes the function to be called by the qt
    main thread rather than synchronously...

    NOTE: This makes the call async handled by the qt main
    loop code.  you can NOT return anything.
    '''
    def idleadd(*args, **kwargs):

        if iAmQtSafeThread():
            return func(*args, **kwargs)

        guiq.append( (func, args, kwargs) )

    functools.update_wrapper(idleadd,func)
    return idleadd

def workthread(func):
    '''
    Proxy a call through the single vqt.main worker thread
    (who exists to keep the GUI from blocking on stuff... )
    '''
    # If we're already the work thread, just do it.
    def workadd(*args, **kwargs):
        if getattr(current_thread(),'VQtWorkerThread',False):
            return func(*args, **kwargs)

        workerq.append( (func,args,kwargs) )

    functools.update_wrapper(workadd,func)
    return workadd

def boredthread(func):
    '''
    The same as "workthread" above, but drop the request on the
    floor if the worker thread already has better things to do...
    '''
    # If we're already the work thread, just do it.
    def workadd(*args, **kwargs):
        if getattr(current_thread(),'VQtWorkerThread',False):
            return func(*args, **kwargs)

        if not len(workerq):
            workerq.append( (func,args,kwargs) )

    functools.update_wrapper(workadd,func)
    return workadd

def idlethreadsync(func):
    '''
    Similar to idlethread except that it is synchronous and able
    to return values.
    '''
    q = Queue()
    def dowork(*args, **kwargs):
        try:
            q.put(func(*args, **kwargs))
        except Exception as e:
            q.put(e)

    def idleadd(*args, **kwargs):
        if iAmQtSafeThread():
            return func(*args, **kwargs)

        guiq.append( (dowork, args, kwargs) )
        return q.get()

    functools.update_wrapper(idleadd,func)
    return idleadd

class QFireThread(QtCore.QThread):
    def __init__(self, callable, args, kwargs):
        QtCore.QThread.__init__(self)
        self.args = args
        self.kwargs = kwargs
        self.callable = callable

    def run(self):
        self.callable(*self.args, **self.kwargs)

def fireqtthread(func):

    def doqtthread(*args, **kwargs):
        func._qt_thread = QFireThread(func, args, kwargs)
        func._qt_thread.start()
        return func._qt_thread

    functools.update_wrapper(doqtthread,func)
    return doqtthread

def iAmQtSafeThread():
    return getattr(current_thread(),'QtSafeThread',False)

class QEventThread(QtCore.QThread):
    '''
    A thread who exists to consume callback requests from the
    given workq and fire them into Qt *safely*.
    '''
    idleadd = QtCore.pyqtSignal(object, object, object)

    def __init__(self, workq):
        QtCore.QThread.__init__(self)
        self.workq = workq

    def run(self):
        while True:
            try:

                todo = self.workq.get()
                if todo is None:
                    continue

                func, args, kwargs = todo
                if func is None:
                    return

                self.idleadd.emit(func,args,kwargs)

            except Exception as e:
                logger.warning('vqt event thread: %s', str(e))

class VQApplication(QApplication):

    guievents = QtCore.pyqtSignal(str,object)

    def __init__(self, *args, **kwargs):
        QApplication.__init__(self, *args, **kwargs)
        self.vqtchans = {}

    def callFromQtLoop(self, callback, args, kwargs):
        callback(*args,**kwargs)


class QEventChannel(QtCore.QObject):
    guievents = QtCore.pyqtSignal(str,object)


@e_threads.firethread
def workerThread():
    # We are *not* allowed to make Qt API calls
    current_thread().VQtWorkerThread = True
    while True:
        try:
            todo = workerq.get()
            if todo is not None:
                func, args, kwargs = todo

                if func is None:
                    return

                try:
                    func(*args, **kwargs)
                except:
                    sys.excepthook(*sys.exc_info())

        except Exception as e:
            logger.warning('vqt worker warning: %s', str(e))

def isGuiStarted():
    return QApplication.instance() is not None

def startup(css=None):
    # yea yea.... globals suck...
    global qapp     # the main QApplication
    global guiq     # queue of GUI calls to proxy
    global ethread  # QtThread that consumes guiq
    global workerq  # queue of "worker" calls to proxy

    # skip startup if already started
    if isGuiStarted():
        print("Returning from startup... already QApplication running")
        return

    guiq = e_threads.EnviQueue()
    workerq = e_threads.EnviQueue()

    current_thread().QtSafeThread = True
    qapp = VQApplication(sys.argv)
    if css:
        qapp.setStyleSheet( css )

    ethread = QEventThread(guiq)
    ethread.idleadd.connect( qapp.callFromQtLoop )
    ethread.start()

    workerThread()

def main():
    global qapp

    if not iAmQtSafeThread():
        raise Exception('main() must be called by same thread as startup()!')

    qapp.exec_()

def eatevents():
    global qapp
    qapp.processEvents()

def vqtevent(event,einfo):
    '''
    Fire an event into the application wide GUI events subsystem.
    Each event should be an event name ( str ) and arbitrary event
    info context.
    '''
    global qapp
    qapp.guievents.emit(event, einfo)
    chan = qapp.vqtchans.get(event)
    if chan is not None:
        chan.guievents.emit(event, einfo)

def vqtconnect(callback, event=None):
    '''
    Connect to the application wide "gui events" which has
    a callback syntax:
        callback(event,einfo)

    Optionally specify an event name to only recieve events
    of the specified type.
    '''
    global qapp
    if event is None:
        qapp.guievents.connect(callback)
        return

    chan = qapp.vqtchans.get(event)
    if chan is None:
        chan = QEventChannel()
        qapp.vqtchans[event] = chan

    chan.guievents.connect(callback)

def vqtdisconnect(callback, event=None):
    '''
    Connect to the application wide "gui events" which has
    a callback syntax:
        callback(event,einfo)

    Optionally specify an event name to only recieve events
    of the specified type.
    '''
    global qapp
    if event is None:
        qapp.guievents.disconnect( callback )
        return

    chan = qapp.vqtchans.get(event)
    if chan is not None:
        chan.guievents.disconnect(callback)

def getOpenFileName(*args, **kwargs):
        fname = QFileDialog.getOpenFileName(*args, **kwargs)
        if type(fname) == tuple:
            return fname[0]
        return fname

def getSaveFileName(*args, **kwargs):
        fname = QFileDialog.getSaveFileName(*args, **kwargs)
        if type(fname) == tuple:
            return fname[0]
        return fname





'''
Generates HTML documentation for Vdb and related APIs using epydoc.

Command line options (all are optional):

    --dotpath=PATH          To generate class heirarchy graphs as part of
                            documentation, specify the path to the GraphViz
                            dot executable

    -o PATH, --output=PATH  The output directory.  If PATH does not exist,
                            then it will be created.  If it exists and is
                            non-empty, then the current directory will be
                            renamed and a new one created.

    NAMES...                If you have additional files / modules to be
                            included in the documentation, specify them on the
                            command line after other options.

Usage Examples: (from command prompt in directory containing vdb directory)
  python -m vdb.tools.gendocs
  python -m vdb.tools.gendocs --dotpath "D:\\downloads\\Graphviz 2.28\\bin\\dot.exe" ..\\vdb-extensions-git\\*.py
  python -m vdb.tools.gendocs --dotpath "D:\\downloads\\Graphviz 2.28\\bin\\dot.exe" --output mydocs ..\\vdb-extensions-git\\*.py
'''
import os
import sys
import datetime

import epydoc
import epydoc.cli


def vdbEpydoc():
    '''
    Generates HTML documentation for Vdb and related APIs using epydoc.
    '''

    if len(sys.argv) > 1 and (sys.argv[1] == '-h' or sys.argv[1] == '--help'):
        print(__doc__)
        return

    toolsdirname = os.path.dirname(os.path.abspath(__file__))

    outputdir = os.path.join(toolsdirname, os.path.join('..', os.path.join('..', 'apidocs')))
    configfile = os.path.join(toolsdirname, 'vdb.epydoc')
    cssfile = os.path.join(toolsdirname, 'vdb.epydoc.css')

    newargv = ['epydoc', '--config', configfile, '--output', outputdir, '--css', cssfile]

    if sys.argv > 1:
        newargv += sys.argv[1:]

    # Create set the fake command line arguments so they will be parsed
    sys.argv = newargv

    options, names = epydoc.cli.parse_arguments()

    # rename the output directory if it exists to give clean output
    print(options.target)
    if os.path.exists(options.target):
        # path exists, verify it is a directory
        if not os.path.isdir(options.target):
            print('%s exists but is not a directory, exiting', options.target)
            return

        # if dir is not empty, rename it.
        if len(os.listdir(options.target)) > 0:
            now = datetime.datetime.now()
            targetdirname = os.path.join(options.target, '-pre-' + now.strftime('%Y-%m-%d_%H%M%S'))

            print('output dir: %s exists and is not empty. moving old dir to name %s' % (options.target, targetdirname))
            os.rename(options.target, targetdirname)

    epydoc.cli.main(options, names)


if __name__ == '__main__':
    vdbEpydoc()



import vivisect.exc as v_exc
import vivisect.impemu.monitor as viv_monitor

import envi
import envi.archs.amd64 as e_amd64

from vivisect.const import *

import vivisect.analysis.generic.switchcase as vag_switch

regops = set(['cmp', 'sub'])

class AnalysisMonitor(viv_monitor.AnalysisMonitor):

    def __init__(self, vw, fva):
        viv_monitor.AnalysisMonitor.__init__(self, vw, fva)
        self.retbytes = None
        self.addDynamicBranchHandler(vag_switch.analyzeJmp)
        self.badops = vw.arch.archGetBadOps()

    def prehook(self, emu, op, starteip):

        if op in self.badops:
            raise v_exc.BadOpBytes(op.va)

        viv_monitor.AnalysisMonitor.prehook(self, emu, op, starteip)

        if op.iflags & envi.IF_RET:
            if len(op.opers):
                self.retbytes = op.opers[0].imm

##### FIXME: this should be all done through the calling convention
sysvamd64argnames = {
    0: ('rdi', e_amd64.REG_RDI),
    1: ('rsi', e_amd64.REG_RSI),
    2: ('rdx', e_amd64.REG_RDX),
    3: ('rcx', e_amd64.REG_RCX),
    4: ('r8',  e_amd64.REG_R8),
    5: ('r9',  e_amd64.REG_R9),
}


msx64argnames = {
    0: ('rcx', e_amd64.REG_RCX),
    1: ('rdx', e_amd64.REG_RDX),
    2: ('r8',  e_amd64.REG_R8),
    3: ('r9',  e_amd64.REG_R9),
}


arch_bindings = {
    'msx64call': msx64argnames,
    'sysvamd64call': sysvamd64argnames,
    None: [],
}


def getName(names, idx):
    ret = names.get(idx)
    if ret is None:
        name = 'arg%d' % idx
    else:
        name, idx = ret
    return name


def buildFunctionApi(vw, fva, emu, emumon):
    argc = 0
    funcargs = []
    callconv = vw.getMeta('DefaultCall')
    argnames = arch_bindings.get(callconv)
    undefregs = set(emu.getUninitRegUse())

    for argnum in range(len(argnames), 0, -1):
        argname, argid = argnames[argnum-1]
        if argid in undefregs:
            argc = argnum
            break

    if callconv == 'msx64call':
        # For msx64call there's the shadow space..
        if emumon.stackmax >= 40:
            targc = (emumon.stackmax >> 3) - 1
            if targc > 40:
                emumon.logAnomaly(emu, fva, 'Crazy Stack Offset Touched: 0x%.8x' % emumon.stackmax)
            else:
                argc = targc

        # Add the shadow space "locals"
        vw.setFunctionLocal(fva, 8,  LSYM_NAME, ('void *', 'shadow0'))
        vw.setFunctionLocal(fva, 16, LSYM_NAME, ('void *', 'shadow1'))
        vw.setFunctionLocal(fva, 24, LSYM_NAME, ('void *', 'shadow2'))
        vw.setFunctionLocal(fva, 32, LSYM_NAME, ('void *', 'shadow3'))

    elif callconv == 'sysvamd64call':
        if emumon.stackmax > 0:
            targc = (emumon.stackmax >> 3) + 6
            if targc > 40:
                emumon.logAnomaly(emu, fva, 'Crazy Stack Offset Touched: 0x%.8x' % emumon.stackmax)
            else:
                argc = targc

    funcargs = [('int', getName(argnames, i)) for i in range(argc)]
    api = ('int', None, callconv, None, funcargs)
    vw.setFunctionApi(fva, api)
    return api


def analyzeFunction(vw, fva):

    emu = vw.getEmulator(va=fva)
    emumon = AnalysisMonitor(vw, fva)

    emu.setEmulationMonitor(emumon)
    emu.runFunction(fva, maxhit=1)

    # Do we already have API info in meta?
    # NOTE: do *not* use getFunctionApi here, it will make one!
    api = vw.getFunctionMeta(fva, 'api')
    if api is None:
        api = buildFunctionApi(vw, fva, emu, emumon)

    rettype, retname, callconv, callname, callargs = api

    argc = len(callargs)
    cc = emu.getCallingConvention(callconv)
    stcount = cc.getNumStackArgs(emu, argc)
    stackidx = argc - stcount
    baseoff = cc.getStackArgOffset(emu, argc)

    # Register our stack args as function locals
    for i in range(stcount):
        vw.setFunctionLocal(fva, baseoff + (i * 8), LSYM_FARG, i+stackidx)

    emumon.addAnalysisResults(vw, emu)



# APIs for Windows 32-bit libraries ws2_32, mswsock, wsock32, and wininet.
# Format:  rettype, retname, callconv, exactname, arglist(type, name)
#          arglist type is one of ['int', 'void *']
#          arglist name is one of [None, 'funcptr', 'obj', 'ptr']

api_defs = {
    # WS2_32
    'ws2_32.main_entry':( 'int', None, 'stdcall', 'ws2_32.main_entry', (('int', None), ('int', None), ('int', None)) ),
    'ws2_32.freeaddrinfow':( 'int', None, 'stdcall', 'ws2_32.FreeAddrInfoW', (('void *', 'ptr'),) ),
    'ws2_32.getaddrinfow':( 'int', None, 'stdcall', 'ws2_32.GetAddrInfoW', (('void *', 'ptr'), ('int', None), ('int', None), ('int', None)) ),
    'ws2_32.getnameinfow':( 'int', None, 'stdcall', 'ws2_32.GetNameInfoW', (('int', None), ('int', None), ('void *', 'ptr'), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'ws2_32.wep':( 'int', None, 'cdecl', 'ws2_32.WEP', () ),
    'ws2_32.wpucompleteoverlappedrequest':( 'int', None, 'stdcall', 'ws2_32.WPUCompleteOverlappedRequest', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'ws2_32.wsaaccept':( 'int', None, 'stdcall', 'ws2_32.WSAAccept', ( ('void *', 'obj'), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'ws2_32.wsaaddresstostringa':( 'int', None, 'stdcall', 'ws2_32.WSAAddressToStringA', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'ws2_32.wsaaddresstostringw':( 'int', None, 'stdcall', 'ws2_32.WSAAddressToStringW', (('void *', 'ptr'), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'ws2_32.wsaasyncgethostbyaddr':( 'int', None, 'stdcall', 'ws2_32.WSAAsyncGetHostByAddr', (('int', None), ('int', None), ('int', None), ('DWORD', None), ('int', None), ('int', None), ('int', None)) ),
    'ws2_32.wsaasyncgethostbyname':( 'int', None, 'stdcall', 'ws2_32.WSAAsyncGetHostByName', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'ws2_32.wsaasyncgetprotobyname':( 'int', None, 'stdcall', 'ws2_32.WSAAsyncGetProtoByName', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'ws2_32.wsaasyncgetprotobynumber':( 'int', None, 'stdcall', 'ws2_32.WSAAsyncGetProtoByNumber', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'ws2_32.wsaasyncgetservbyname':( 'int', None, 'stdcall', 'ws2_32.WSAAsyncGetServByName', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'ws2_32.wsaasyncgetservbyport':( 'int', None, 'stdcall', 'ws2_32.WSAAsyncGetServByPort', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'ws2_32.wsaasyncselect':( 'int', None, 'stdcall', 'ws2_32.WSAAsyncSelect', ( ('int', None), ('int', None), ('int', None)) ),
    'ws2_32.wsacancelasyncrequest':( 'int', None, 'stdcall', 'ws2_32.WSACancelAsyncRequest', (('int', None),) ),
    'ws2_32.wsacancelblockingcall':( 'int', None, 'stdcall', 'ws2_32.WSACancelBlockingCall', () ),
    'ws2_32.wsacleanup':( 'int', None, 'stdcall', 'ws2_32.WSACleanup', () ),
    'ws2_32.wsacloseevent':( 'int', None, 'stdcall', 'ws2_32.WSACloseEvent', (('HANDLE', None),) ),
    'ws2_32.wsaconnect':( 'int', None, 'stdcall', 'ws2_32.WSAConnect', ( ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'ws2_32.wsacreateevent':( 'int', None, 'cdecl', 'ws2_32.WSACreateEvent', () ),
    'ws2_32.wsaduplicatesocketa':( 'int', None, 'stdcall', 'ws2_32.WSADuplicateSocketA', ( ('int', None), ('int', None)) ),
    'ws2_32.wsaduplicatesocketw':( 'int', None, 'stdcall', 'ws2_32.WSADuplicateSocketW', ( ('int', None), ('void *', 'ptr')) ),
    'ws2_32.wsaenumnamespaceprovidersa':( 'int', None, 'stdcall', 'ws2_32.WSAEnumNameSpaceProvidersA', (('int', None), ('int', None)) ),
    'ws2_32.wsaenumnamespaceprovidersw':( 'int', None, 'stdcall', 'ws2_32.WSAEnumNameSpaceProvidersW', (('int', None), ('int', None)) ),
    'ws2_32.wsaenumnetworkevents':( 'int', None, 'stdcall', 'ws2_32.WSAEnumNetworkEvents', ( ('int', None), ('int', None)) ),
    'ws2_32.wsaenumprotocolsa':( 'int', None, 'stdcall', 'ws2_32.WSAEnumProtocolsA', (('int', None), ('int', None), ('int', None)) ),
    'ws2_32.wsaenumprotocolsw':( 'int', None, 'stdcall', 'ws2_32.WSAEnumProtocolsW', (('int', None), ('int', None), ('void *', 'ptr')) ),
    'ws2_32.wsaeventselect':( 'int', None, 'stdcall', 'ws2_32.WSAEventSelect', ( ('int', None), ('int', None)) ),
    'ws2_32.wsagetlasterror':( 'int', None, 'stdcall', 'ws2_32.WSAGetLastError', () ),
    'ws2_32.wsagetoverlappedresult':( 'int', None, 'stdcall', 'ws2_32.WSAGetOverlappedResult', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'ws2_32.wsagetqosbyname':( 'int', None, 'stdcall', 'ws2_32.WSAGetQOSByName', ( ('int', None), ('int', None)) ),
    'ws2_32.wsagetserviceclassinfoa':( 'int', None, 'stdcall', 'ws2_32.WSAGetServiceClassInfoA', (('int', None), ('int', None), ('int', None), ('int', None)) ),
    'ws2_32.wsagetserviceclassinfow':( 'int', None, 'stdcall', 'ws2_32.WSAGetServiceClassInfoW', (('int', None), ('int', None), ('int', None), ('int', None)) ),
    'ws2_32.wsagetserviceclassnamebyclassida':( 'int', None, 'stdcall', 'ws2_32.WSAGetServiceClassNameByClassIdA', (('int', None), ('int', None), ('int', None)) ),
    'ws2_32.wsagetserviceclassnamebyclassidw':( 'int', None, 'stdcall', 'ws2_32.WSAGetServiceClassNameByClassIdW', (('int', None), ('int', None), ('int', None)) ),
    'ws2_32.wsahtonl':( 'int', None, 'stdcall', 'ws2_32.WSAHtonl', ( ('int', None), ('int', None)) ),
    'ws2_32.wsahtons':( 'int', None, 'stdcall', 'ws2_32.WSAHtons', ( ('int', None), ('int', None)) ),
    'ws2_32.wsainstallserviceclassa':( 'int', None, 'stdcall', 'ws2_32.WSAInstallServiceClassA', (('int', None),) ),
    'ws2_32.wsainstallserviceclassw':( 'int', None, 'stdcall', 'ws2_32.WSAInstallServiceClassW', (('int', None),) ),
    'ws2_32.wsaioctl':( 'int', None, 'stdcall', 'ws2_32.WSAIoctl', ( ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('void *', 'ptr'), ('int', None), ('int', None)) ),
    'ws2_32.wsaisblocking':( 'int', None, 'stdcall', 'ws2_32.WSAIsBlocking', () ),
    'ws2_32.wsajoinleaf':( 'int', None, 'stdcall', 'ws2_32.WSAJoinLeaf', ( ('void *', 'obj'), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'ws2_32.wsalookupservicebegina':( 'int', None, 'stdcall', 'ws2_32.WSALookupServiceBeginA', (('int', None), ('int', None), ('void *', 'ptr')) ),
    'ws2_32.wsalookupservicebeginw':( 'int', None, 'stdcall', 'ws2_32.WSALookupServiceBeginW', ( ('void *', 'ptr'), ('int', None), ('void *', 'ptr')) ),
    'ws2_32.wsalookupserviceend':( 'int', None, 'stdcall', 'ws2_32.WSALookupServiceEnd', () ),
    'ws2_32.wsalookupservicenexta':( 'int', None, 'stdcall', 'ws2_32.WSALookupServiceNextA', ( ('int', None), ('int', None), ('int', None)) ),
    'ws2_32.wsalookupservicenextw':( 'int', None, 'stdcall', 'ws2_32.WSALookupServiceNextW', ( ('int', None), ('void *', 'ptr'), ('void *', 'ptr')) ),
    'ws2_32.wsanspioctl':( 'int', None, 'stdcall', 'ws2_32.WSANSPIoctl', ( ('void *', 'obj'), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'ws2_32.wsantohl':( 'int', None, 'stdcall', 'ws2_32.WSANtohl', ( ('int', None), ('int', None)) ),
    'ws2_32.wsantohs':( 'int', None, 'stdcall', 'ws2_32.WSANtohs', ( ('int', None), ('int', None)) ),
    'ws2_32.wsaproviderconfigchange':( 'int', None, 'stdcall', 'ws2_32.WSAProviderConfigChange', (('int', None), ('int', None), ('int', None)) ),
    'ws2_32.wsarecv':( 'int', None, 'stdcall', 'ws2_32.WSARecv', ( ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'ws2_32.wsarecvdisconnect':( 'int', None, 'stdcall', 'ws2_32.WSARecvDisconnect', ( ('int', None),) ),
    'ws2_32.wsarecvfrom':( 'int', None, 'stdcall', 'ws2_32.WSARecvFrom', ( ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'ws2_32.wsaremoveserviceclass':( 'int', None, 'stdcall', 'ws2_32.WSARemoveServiceClass', (('int', None),) ),
    'ws2_32.wsaresetevent':( 'int', None, 'cdecl', 'ws2_32.WSAResetEvent', () ),
    'ws2_32.wsasend':( 'int', None, 'stdcall', 'ws2_32.WSASend', ( ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'ws2_32.wsasenddisconnect':( 'int', None, 'stdcall', 'ws2_32.WSASendDisconnect', ( ('int', None),) ),
    'ws2_32.wsasendto':( 'int', None, 'stdcall', 'ws2_32.WSASendTo', ( ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'ws2_32.wsasetblockinghook':( 'int', None, 'stdcall', 'ws2_32.WSASetBlockingHook', ( ('int', None),) ),
    'ws2_32.wsasetevent':( 'int', None, 'cdecl', 'ws2_32.WSASetEvent', () ),
    'ws2_32.wsasetlasterror':( 'int', None, 'cdecl', 'ws2_32.WSASetLastError', () ),
    'ws2_32.wsasetservicea':( 'int', None, 'stdcall', 'ws2_32.WSASetServiceA', (('int', None), ('int', None), ('int', None)) ),
    'ws2_32.wsasetservicew':( 'int', None, 'stdcall', 'ws2_32.WSASetServiceW', (('int', None), ('int', None), ('int', None)) ),
    'ws2_32.wsasocketa':( 'int', None, 'stdcall', 'ws2_32.WSASocketA', (('int', None), ('int', None), ('void *', 'obj'), ('int', None), ('int', None), ('int', None)) ),
    'ws2_32.wsasocketw':( 'int', None, 'stdcall', 'ws2_32.WSASocketW', (('int', None), ('int', None), ('void *', 'obj'), ('void *', 'ptr'), ('int', None), ('int', None)) ),
    'ws2_32.wsastartup':( 'int', None, 'stdcall', 'ws2_32.WSAStartup', (('int', None), ('void *', 'ptr')) ),
    'ws2_32.wsastringtoaddressa':( 'int', None, 'stdcall', 'ws2_32.WSAStringToAddressA', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'ws2_32.wsastringtoaddressw':( 'int', None, 'stdcall', 'ws2_32.WSAStringToAddressW', (('int', None), ('int', None), ('int', None), ('void *', 'ptr'), ('void *', 'ptr')) ),
    'ws2_32.wsaunhookblockinghook':( 'int', None, 'stdcall', 'ws2_32.WSAUnhookBlockingHook', () ),
    'ws2_32.wsawaitformultipleevents':( 'int', None, 'cdecl', 'ws2_32.WSAWaitForMultipleEvents', () ),
    'ws2_32.wsapsetpostroutine':( 'int', None, 'stdcall', 'ws2_32.WSApSetPostRoutine', (('int', None),) ),
    'ws2_32.wscdeinstallprovider':( 'int', None, 'stdcall', 'ws2_32.WSCDeinstallProvider', (('int', None), ('int', None)) ),
    'ws2_32.wscenablensprovider':( 'int', None, 'stdcall', 'ws2_32.WSCEnableNSProvider', (('int', None), ('int', None)) ),
    'ws2_32.wscenumprotocols':( 'int', None, 'stdcall', 'ws2_32.WSCEnumProtocols', (('int', None), ('int', None), ('int', None), ('int', None)) ),
    'ws2_32.wscgetproviderpath':( 'int', None, 'stdcall', 'ws2_32.WSCGetProviderPath', (('int', None), ('int', None), ('int', None), ('int', None)) ),
    'ws2_32.wscinstallnamespace':( 'int', None, 'stdcall', 'ws2_32.WSCInstallNameSpace', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'ws2_32.wscinstallprovider':( 'int', None, 'stdcall', 'ws2_32.WSCInstallProvider', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'ws2_32.wscuninstallnamespace':( 'int', None, 'stdcall', 'ws2_32.WSCUnInstallNameSpace', (('int', None),) ),
    'ws2_32.wscupdateprovider':( 'int', None, 'stdcall', 'ws2_32.WSCUpdateProvider', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'ws2_32.wscwritenamespaceorder':( 'int', None, 'stdcall', 'ws2_32.WSCWriteNameSpaceOrder', (('int', None), ('int', None)) ),
    'ws2_32.wscwriteproviderorder':( 'int', None, 'stdcall', 'ws2_32.WSCWriteProviderOrder', (('int', None), ('int', None)) ),
    'ws2_32.__wsafdisset':( 'int', None, 'stdcall', 'ws2_32.__WSAFDIsSet', (('int', None), ('int', None)) ),
    'ws2_32.accept':( 'int', None, 'stdcall', 'ws2_32.accept', ( ('void *', 'obj'), ('int', None), ('int', None)) ),
    'ws2_32.bind':( 'int', None, 'stdcall', 'ws2_32.bind', ( ('int', None), ('int', None), ('int', None)) ),
    'ws2_32.closesocket':( 'int', None, 'stdcall', 'ws2_32.closesocket', (('int', None),) ),
    'ws2_32.connect':( 'int', None, 'stdcall', 'ws2_32.connect', ( ('int', None), ('int', None), ('int', None)) ),
    'ws2_32.freeaddrinfo':( 'int', None, 'stdcall', 'ws2_32.freeaddrinfo', (('void *', 'ptr'),) ),
    'ws2_32.getaddrinfo':( 'int', None, 'stdcall', 'ws2_32.getaddrinfo', (('int', None), ('int', None), ('int', None), ('int', None)) ),
    'ws2_32.gethostbyaddr':( 'int', None, 'stdcall', 'ws2_32.gethostbyaddr', ( ('int', None), ('int', None), ('int', None)) ),
    'ws2_32.gethostbyname':( 'int', None, 'stdcall', 'ws2_32.gethostbyname', ( ('void *', 'ptr'),) ),
    'ws2_32.gethostname':( 'int', None, 'stdcall', 'ws2_32.gethostname', ( ('void *', 'ptr'), ('int', None)) ),
    'ws2_32.getnameinfo':( 'int', None, 'stdcall', 'ws2_32.getnameinfo', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'ws2_32.getpeername':( 'int', None, 'stdcall', 'ws2_32.getpeername', ( ('int', None), ('int', None)) ),
    'ws2_32.getprotobyname':( 'int', None, 'stdcall', 'ws2_32.getprotobyname', (('int', None),) ),
    'ws2_32.getprotobynumber':( 'int', None, 'stdcall', 'ws2_32.getprotobynumber', ( ('int', None),) ),
    'ws2_32.getservbyname':( 'int', None, 'stdcall', 'ws2_32.getservbyname', (('void *', 'ptr'), ('int', None)) ),
    'ws2_32.getservbyport':( 'int', None, 'stdcall', 'ws2_32.getservbyport', (('int', None), ('int', None)) ),
    'ws2_32.getsockname':( 'int', None, 'stdcall', 'ws2_32.getsockname', ( ('int', None), ('int', None)) ),
    'ws2_32.getsockopt':( 'int', None, 'stdcall', 'ws2_32.getsockopt', ( ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'ws2_32.htonl':( 'int', None, 'stdcall', 'ws2_32.htonl', (('int', None),) ),
    'ws2_32.htons':( 'int', None, 'stdcall', 'ws2_32.htons', (('int', None),) ),
    'ws2_32.inet_addr':( 'int', None, 'stdcall', 'ws2_32.inet_addr', (('int', None),) ),
    'ws2_32.inet_ntoa':( 'int', None, 'stdcall', 'ws2_32.inet_ntoa', ( ('int', None),) ),
    'ws2_32.ioctlsocket':( 'int', None, 'stdcall', 'ws2_32.ioctlsocket', ( ('int', None), ('int', None)) ),
    'ws2_32.listen':( 'int', None, 'stdcall', 'ws2_32.listen', ( ('int', None),) ),
    'ws2_32.ntohl':( 'int', None, 'stdcall', 'ws2_32.ntohl', (('int', None),) ),
    'ws2_32.ntohs':( 'int', None, 'stdcall', 'ws2_32.ntohs', (('int', None),) ),
    'ws2_32.recv':( 'int', None, 'stdcall', 'ws2_32.recv', ( ('int', None), ('int', None), ('int', None)) ),
    'ws2_32.recvfrom':( 'int', None, 'stdcall', 'ws2_32.recvfrom', ( ('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'ws2_32.select':( 'int', None, 'stdcall', 'ws2_32.select', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'ws2_32.send':( 'int', None, 'stdcall', 'ws2_32.send', ( ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'ws2_32.sendto':( 'int', None, 'stdcall', 'ws2_32.sendto', ( ('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'ws2_32.setsockopt':( 'int', None, 'stdcall', 'ws2_32.setsockopt', ( ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'ws2_32.shutdown':( 'int', None, 'stdcall', 'ws2_32.shutdown', ( ('int', None), ('int', None)) ),
    'ws2_32.socket':( 'int', None, 'stdcall', 'ws2_32.socket', ( ('int', None), ('int', None), ('void *', 'obj')) ),

    # MSWSOCK
    'mswsock.main_entry':( 'int', None, 'stdcall', 'mswsock.main_entry', (('int', None), ('int', None), ('int', None)) ),
    'mswsock.acceptex':( 'int', None, 'stdcall', 'mswsock.AcceptEx', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'mswsock.enumprotocolsa':( 'int', None, 'stdcall', 'mswsock.EnumProtocolsA', (('int', None), ('int', None), ('void *', 'ptr')) ),
    'mswsock.enumprotocolsw':( 'int', None, 'stdcall', 'mswsock.EnumProtocolsW', (('int', None), ('int', None), ('void *', 'ptr')) ),
    'mswsock.getacceptexsockaddrs':( 'int', None, 'stdcall', 'mswsock.GetAcceptExSockaddrs', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'mswsock.getaddressbynamea':( 'int', None, 'stdcall', 'mswsock.GetAddressByNameA', ( ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'mswsock.getaddressbynamew':( 'int', None, 'stdcall', 'mswsock.GetAddressByNameW', ( ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('void *', 'ptr')) ),
    'mswsock.getnamebytypea':( 'int', None, 'stdcall', 'mswsock.GetNameByTypeA', (('int', None), ('int', None), ('int', None)) ),
    'mswsock.getnamebytypew':( 'int', None, 'stdcall', 'mswsock.GetNameByTypeW', (('int', None), ('void *', 'ptr'), ('int', None)) ),
    'mswsock.getservicea':( 'int', None, 'stdcall', 'mswsock.GetServiceA', ( ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'mswsock.getservicew':( 'int', None, 'stdcall', 'mswsock.GetServiceW', ( ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'mswsock.gettypebynamea':( 'int', None, 'stdcall', 'mswsock.GetTypeByNameA', (('int', None), ('void *', 'ptr')) ),
    'mswsock.gettypebynamew':( 'int', None, 'stdcall', 'mswsock.GetTypeByNameW', (('void *', 'ptr'), ('void *', 'ptr')) ),
    'mswsock.migratewinsockconfiguration':( 'int', None, 'stdcall', 'mswsock.MigrateWinsockConfiguration', (('int', None), ('void *', 'funcptr'), ('int', None)) ),
    'mswsock.nploadnamespaces':( 'int', None, 'stdcall', 'mswsock.NPLoadNameSpaces', (('int', None), ('int', None), ('int', None)) ),
    'mswsock.nspstartup':( 'int', None, 'stdcall', 'mswsock.NSPStartup', (('int', None), ('int', None)) ),
    'mswsock.servicemain':( 'int', None, 'stdcall', 'mswsock.ServiceMain', (('int', None), ('int', None)) ),
    'mswsock.setservicea':( 'int', None, 'stdcall', 'mswsock.SetServiceA', ( ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'mswsock.setservicew':( 'int', None, 'stdcall', 'mswsock.SetServiceW', ( ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'mswsock.startwsdpservice':( 'int', None, 'cdecl', 'mswsock.StartWsdpService', () ),
    'mswsock.stopwsdpservice':( 'int', None, 'cdecl', 'mswsock.StopWsdpService', () ),
    'mswsock.svchostpushserviceglobals':( 'int', None, 'stdcall', 'mswsock.SvchostPushServiceGlobals', (('int', None),) ),
    'mswsock.transmitfile':( 'int', None, 'stdcall', 'mswsock.TransmitFile', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'mswsock.wsarecvex':( 'int', None, 'stdcall', 'mswsock.WSARecvEx', (('int', None), ('int', None), ('int', None), ('int', None)) ),
    'mswsock.wspstartup':( 'int', None, 'stdcall', 'mswsock.WSPStartup', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'mswsock.dn_expand':( 'int', None, 'stdcall', 'mswsock.dn_expand', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'mswsock.getnetbyname':( 'int', None, 'stdcall', 'mswsock.getnetbyname', (('int', None),) ),
    'mswsock.inet_network':( 'int', None, 'stdcall', 'mswsock.inet_network', (('int', None),) ),
    'mswsock.rcmd':( 'int', None, 'stdcall', 'mswsock.rcmd', ( ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'mswsock.rexec':( 'int', None, 'stdcall', 'mswsock.rexec', ( ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'mswsock.rresvport':( 'int', None, 'stdcall', 'mswsock.rresvport', (('void *', 'ptr'),) ),
    'mswsock.s_perror':( 'int', None, 'stdcall', 'mswsock.s_perror', (('void *', 'ptr'), ('int', None)) ),
    'mswsock.sethostname':( 'int', None, 'stdcall', 'mswsock.sethostname', (('int', None), ('int', None)) ),

    # WSOCK32
    'wsock32.main_entry':( 'int', None, 'stdcall', 'wsock32.main_entry', (('int', None), ('int', None), ('int', None)) ),
    'wsock32.getsockopt':( 'int', None, 'stdcall', 'wsock32.getsockopt', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wsock32.recv':( 'int', None, 'stdcall', 'wsock32.recv', (('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wsock32.recvfrom':( 'int', None, 'stdcall', 'wsock32.recvfrom', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wsock32.setsockopt':( 'int', None, 'stdcall', 'wsock32.setsockopt', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wsock32.acceptex':('int', None, 'stdcall', 'mswsock.AcceptEx', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None))),
    'wsock32.enumprotocolsa':('int', None, 'stdcall', 'mswsock.EnumProtocolsA', (('int', None), ('int', None), ('void *', 'ptr'))),
    'wsock32.enumprotocolsw':('int', None, 'stdcall', 'mswsock.EnumProtocolsW', (('int', None), ('int', None), ('void *', 'ptr'))),
    'wsock32.getacceptexsockaddrs':('int', None, 'stdcall', 'mswsock.GetAcceptExSockaddrs', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None))),
    'wsock32.getaddressbynamea':('int', None, 'stdcall', 'mswsock.GetAddressByNameA', ( ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None))),
    'wsock32.getaddressbynamew':('int', None, 'stdcall', 'mswsock.GetAddressByNameW', ( ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('void *', 'ptr'))),
    'wsock32.getnamebytypea':('int', None, 'stdcall', 'mswsock.GetNameByTypeA', (('int', None), ('int', None), ('int', None))),
    'wsock32.getnamebytypew':('int', None, 'stdcall', 'mswsock.GetNameByTypeW', (('int', None), ('void *', 'ptr'), ('int', None))),
    'wsock32.getservicea':('int', None, 'stdcall', 'mswsock.GetServiceA', ( ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None))),
    'wsock32.getservicew':('int', None, 'stdcall', 'mswsock.GetServiceW', ( ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None))),
    'wsock32.gettypebynamea':('int', None, 'stdcall', 'mswsock.GetTypeByNameA', (('int', None), ('void *', 'ptr'))),
    'wsock32.gettypebynamew':('int', None, 'stdcall', 'mswsock.GetTypeByNameW', (('void *', 'ptr'), ('void *', 'ptr'))),
    'wsock32.migratewinsockconfiguration':('int', None, 'stdcall', 'mswsock.MigrateWinsockConfiguration', (('int', None), ('void *', 'funcptr'), ('int', None))),
    'wsock32.nploadnamespaces':('int', None, 'stdcall', 'mswsock.NPLoadNameSpaces', (('int', None), ('int', None), ('int', None))),
    'wsock32.setservicea':('int', None, 'stdcall', 'mswsock.SetServiceA', ( ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None))),
    'wsock32.setservicew':('int', None, 'stdcall', 'mswsock.SetServiceW', ( ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None))),
    'wsock32.transmitfile':('int', None, 'stdcall', 'mswsock.TransmitFile', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None))),
    'wsock32.wep':('int', None, 'cdecl', 'ws2_32.WEP', ()),
    'wsock32.wsaasyncgethostbyaddr':('int', None, 'stdcall', 'ws2_32.WSAAsyncGetHostByAddr', (('int', None), ('int', None), ('int', None), ('DWORD', None), ('int', None), ('int', None), ('int', None))),
    'wsock32.wsaasyncgethostbyname':('int', None, 'stdcall', 'ws2_32.WSAAsyncGetHostByName', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None))),
    'wsock32.wsaasyncgetprotobyname':('int', None, 'stdcall', 'ws2_32.WSAAsyncGetProtoByName', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None))),
    'wsock32.wsaasyncgetprotobynumber':('int', None, 'stdcall', 'ws2_32.WSAAsyncGetProtoByNumber', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None))),
    'wsock32.wsaasyncgetservbyname':('int', None, 'stdcall', 'ws2_32.WSAAsyncGetServByName', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None))),
    'wsock32.wsaasyncgetservbyport':('int', None, 'stdcall', 'ws2_32.WSAAsyncGetServByPort', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None))),
    'wsock32.wsaasyncselect':('int', None, 'stdcall', 'ws2_32.WSAAsyncSelect', ( ('int', None), ('int', None), ('int', None))),
    'wsock32.wsacancelasyncrequest':('int', None, 'stdcall', 'ws2_32.WSACancelAsyncRequest', (('int', None),)),
    'wsock32.wsacancelblockingcall':('int', None, 'stdcall', 'ws2_32.WSACancelBlockingCall', ()),
    'wsock32.wsacleanup':('int', None, 'stdcall', 'ws2_32.WSACleanup', ()),
    'wsock32.wsagetlasterror':('int', None, 'stdcall', 'ws2_32.WSAGetLastError', ()),
    'wsock32.wsaisblocking':('int', None, 'stdcall', 'ws2_32.WSAIsBlocking', ()),
    'wsock32.wsarecvex':('int', None, 'stdcall', 'mswsock.WSARecvEx', (('int', None), ('int', None), ('int', None), ('int', None))),
    'wsock32.wsasetblockinghook':('int', None, 'stdcall', 'ws2_32.WSASetBlockingHook', ( ('int', None),)),
    'wsock32.wsasetlasterror':('int', None, 'cdecl', 'ws2_32.WSASetLastError', ()),
    'wsock32.wsastartup':('int', None, 'stdcall', 'ws2_32.WSAStartup', (('int', None), ('void *', 'ptr'))),
    'wsock32.wsaunhookblockinghook':('int', None, 'stdcall', 'ws2_32.WSAUnhookBlockingHook', ()),
    'wsock32.wsapsetpostroutine':('int', None, 'stdcall', 'ws2_32.WSApSetPostRoutine', (('int', None),)),
    'wsock32.__wsafdisset':('int', None, 'stdcall', 'ws2_32.__WSAFDIsSet', (('int', None), ('int', None))),
    'wsock32.accept':('int', None, 'stdcall', 'ws2_32.accept', ( ('void *', 'obj'), ('int', None), ('int', None))),
    'wsock32.bind':('int', None, 'stdcall', 'ws2_32.bind', ( ('int', None), ('int', None))),
    'wsock32.closesocket':('int', None, 'stdcall', 'ws2_32.closesocket', ()),
    'wsock32.connect':('int', None, 'stdcall', 'ws2_32.connect', ( ('int', None), ('int', None))),
    'wsock32.dn_expand':('int', None, 'stdcall', 'mswsock.dn_expand', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None))),
    'wsock32.gethostbyaddr':('int', None, 'stdcall', 'ws2_32.gethostbyaddr', ( ('int', None), ('int', None), ('int', None))),
    'wsock32.gethostbyname':('int', None, 'stdcall', 'ws2_32.gethostbyname', ( ('void *', 'ptr'),)),
    'wsock32.gethostname':('int', None, 'stdcall', 'ws2_32.gethostname', ( ('void *', 'ptr'), ('int', None))),
    'wsock32.getnetbyname':('int', None, 'stdcall', 'mswsock.getnetbyname', (('int', None),)),
    'wsock32.getpeername':('int', None, 'stdcall', 'ws2_32.getpeername', ( ('int', None), ('int', None))),
    'wsock32.getprotobyname':('int', None, 'stdcall', 'ws2_32.getprotobyname', (('int', None),)),
    'wsock32.getprotobynumber':('int', None, 'stdcall', 'ws2_32.getprotobynumber', ( ('int', None),)),
    'wsock32.getservbyname':('int', None, 'stdcall', 'ws2_32.getservbyname', (('void *', 'ptr'), ('int', None))),
    'wsock32.getservbyport':('int', None, 'stdcall', 'ws2_32.getservbyport', (('int', None), ('int', None))),
    'wsock32.getsockname':('int', None, 'stdcall', 'ws2_32.getsockname', ( ('int', None), ('int', None))),
    'wsock32.htonl':('int', None, 'stdcall', 'ws2_32.htonl', (('int', None),)),
    'wsock32.htons':('int', None, 'stdcall', 'ws2_32.htons', (('int', None),)),
    'wsock32.inet_addr':('int', None, 'stdcall', 'ws2_32.inet_addr', (('int', None),)),
    'wsock32.inet_network':('int', None, 'stdcall', 'mswsock.inet_network', (('int', None),)),
    'wsock32.inet_ntoa':('int', None, 'stdcall', 'ws2_32.inet_ntoa', ( ('int', None),)),
    'wsock32.ioctlsocket':('int', None, 'stdcall', 'ws2_32.ioctlsocket', ( ('int', None), ('int', None))),
    'wsock32.listen':('int', None, 'stdcall', 'ws2_32.listen', ( ('int', None),)),
    'wsock32.ntohl':('int', None, 'stdcall', 'ws2_32.ntohl', (('int', None),)),
    'wsock32.ntohs':('int', None, 'stdcall', 'ws2_32.ntohs', (('int', None),)),
    'wsock32.rcmd':('int', None, 'stdcall', 'mswsock.rcmd', ( ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None))),
    'wsock32.rexec':('int', None, 'stdcall', 'mswsock.rexec', ( ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None))),
    'wsock32.rresvport':('int', None, 'stdcall', 'mswsock.rresvport', (('void *', 'ptr'),)),
    'wsock32.s_perror':('int', None, 'stdcall', 'mswsock.s_perror', (('void *', 'ptr'), ('int', None))),
    'wsock32.select':('int', None, 'stdcall', 'ws2_32.select', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None))),
    'wsock32.send':('int', None, 'stdcall', 'ws2_32.send', ( ('int', None), ('int', None), ('int', None))),
    'wsock32.sendto':('int', None, 'stdcall', 'ws2_32.sendto', ( ('int', None), ('int', None), ('int', None), ('int', None), ('int', None))),
    'wsock32.sethostname':('int', None, 'stdcall', 'mswsock.sethostname', (('int', None), ('int', None))),
    'wsock32.shutdown':('int', None, 'stdcall', 'ws2_32.shutdown', ( ('int', None),)),
    'wsock32.socket':('int', None, 'stdcall', 'ws2_32.socket', ( ('int', None), ('int', None), ('void *', 'obj'))),

    # WININET
    'wininet.main_entry':( 'int', None, 'stdcall', 'wininet.main_entry', (('int', None), ('int', None), ('int', None)) ),
    'wininet.commiturlcacheentrya':( 'int', None, 'stdcall', 'wininet.CommitUrlCacheEntryA', (('void *', 'ptr'), ('void *', 'ptr'), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('void *', 'ptr'), ('int', None), ('int', None), ('int', None)) ),
    'wininet.commiturlcacheentryw':( 'int', None, 'stdcall', 'wininet.CommitUrlCacheEntryW', (('void *', 'ptr'), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('void *', 'ptr')) ),
    'wininet.createmd5ssohash':( 'int', None, 'stdcall', 'wininet.CreateMD5SSOHash', (('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wininet.createurlcachecontainera':( 'int', None, 'stdcall', 'wininet.CreateUrlCacheContainerA', (('void *', 'ptr'), ('void *', 'ptr'), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wininet.createurlcachecontainerw':( 'int', None, 'stdcall', 'wininet.CreateUrlCacheContainerW', (('int', None), ('int', None), ('void *', 'ptr'), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wininet.createurlcacheentrya':( 'int', None, 'stdcall', 'wininet.CreateUrlCacheEntryA', (('int', None), ('int', None), ('int', None), ('void *', 'ptr'), ('int', None)) ),
    'wininet.createurlcacheentryw':( 'int', None, 'stdcall', 'wininet.CreateUrlCacheEntryW', (('void *', 'ptr'), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wininet.createurlcachegroup':( 'int', None, 'stdcall', 'wininet.CreateUrlCacheGroup', (('int', None), ('int', None)) ),
    'wininet.deleteie3cache':( 'int', None, 'stdcall', 'wininet.DeleteIE3Cache', (('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wininet.deleteurlcachecontainera':( 'int', None, 'stdcall', 'wininet.DeleteUrlCacheContainerA', (('void *', 'ptr'), ('int', None)) ),
    'wininet.deleteurlcachecontainerw':( 'int', None, 'stdcall', 'wininet.DeleteUrlCacheContainerW', (('int', None), ('int', None)) ),
    'wininet.deleteurlcacheentry':( 'int', None, 'stdcall', 'wininet.DeleteUrlCacheEntry', (('int', None),) ),
    'wininet.deleteurlcacheentrya':( 'int', None, 'stdcall', 'wininet.DeleteUrlCacheEntryA', (('int', None),) ),
    'wininet.deleteurlcacheentryw':( 'int', None, 'stdcall', 'wininet.DeleteUrlCacheEntryW', (('void *', 'ptr'),) ),
    'wininet.deleteurlcachegroup':( 'int', None, 'stdcall', 'wininet.DeleteUrlCacheGroup', (('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wininet.deletewpadcachefornetworks':( 'int', None, 'stdcall', 'wininet.DeleteWpadCacheForNetworks', (('int', None),) ),
    'wininet.detectautoproxyurl':( 'int', None, 'stdcall', 'wininet.DetectAutoProxyUrl', ( ('void *', 'ptr'), ('int', None), ('int', None)) ),
    'wininet.dispatchapicall':( 'int', None, 'stdcall', 'wininet.DispatchAPICall', ( ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wininet.dllinstall':( 'int', None, 'stdcall', 'wininet.DllInstall', (('int', None), ('int', None)) ),
    'wininet.findcloseurlcache':( 'int', None, 'stdcall', 'wininet.FindCloseUrlCache', (('int', None),) ),
    'wininet.findfirsturlcachecontainera':( 'int', None, 'stdcall', 'wininet.FindFirstUrlCacheContainerA', (('int', None), ('int', None), ('void *', 'ptr'), ('int', None)) ),
    'wininet.findfirsturlcachecontainerw':( 'int', None, 'stdcall', 'wininet.FindFirstUrlCacheContainerW', (('int', None), ('void *', 'ptr'), ('int', None), ('int', None)) ),
    'wininet.findfirsturlcacheentrya':( 'int', None, 'stdcall', 'wininet.FindFirstUrlCacheEntryA', (('int', None), ('int', None), ('void *', 'ptr')) ),
    'wininet.findfirsturlcacheentryexa':( 'int', None, 'stdcall', 'wininet.FindFirstUrlCacheEntryExA', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('void *', 'ptr'), ('int', None), ('int', None), ('int', None)) ),
    'wininet.findfirsturlcacheentryexw':( 'int', None, 'stdcall', 'wininet.FindFirstUrlCacheEntryExW', (('void *', 'ptr'), ('int', None), ('int', None), ('int', None), ('int', None), ('void *', 'ptr'), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wininet.findfirsturlcacheentryw':( 'int', None, 'stdcall', 'wininet.FindFirstUrlCacheEntryW', (('void *', 'ptr'), ('void *', 'ptr'), ('int', None)) ),
    'wininet.findfirsturlcachegroup':( 'int', None, 'stdcall', 'wininet.FindFirstUrlCacheGroup', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wininet.findnexturlcachecontainera':( 'int', None, 'stdcall', 'wininet.FindNextUrlCacheContainerA', (('int', None), ('int', None), ('void *', 'ptr')) ),
    'wininet.findnexturlcachecontainerw':( 'int', None, 'stdcall', 'wininet.FindNextUrlCacheContainerW', (('int', None), ('void *', 'ptr'), ('int', None)) ),
    'wininet.findnexturlcacheentrya':( 'int', None, 'stdcall', 'wininet.FindNextUrlCacheEntryA', (('int', None), ('int', None), ('void *', 'ptr')) ),
    'wininet.findnexturlcacheentryexa':( 'int', None, 'stdcall', 'wininet.FindNextUrlCacheEntryExA', (('int', None), ('int', None), ('void *', 'ptr'), ('int', None), ('int', None), ('int', None)) ),
    'wininet.findnexturlcacheentryexw':( 'int', None, 'stdcall', 'wininet.FindNextUrlCacheEntryExW', (('int', None), ('void *', 'ptr'), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wininet.findnexturlcacheentryw':( 'int', None, 'stdcall', 'wininet.FindNextUrlCacheEntryW', (('int', None), ('void *', 'ptr'), ('int', None)) ),
    'wininet.findnexturlcachegroup':( 'int', None, 'stdcall', 'wininet.FindNextUrlCacheGroup', (('int', None), ('int', None), ('int', None)) ),
    'wininet.forcenexuslookup':( 'int', None, 'cdecl', 'wininet.ForceNexusLookup', () ),
    'wininet.forcenexuslookupexw':( 'int', None, 'stdcall', 'wininet.ForceNexusLookupExW', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wininet.freeurlcachespacea':( 'int', None, 'stdcall', 'wininet.FreeUrlCacheSpaceA', (('int', None), ('int', None), ('int', None)) ),
    'wininet.freeurlcachespacew':( 'int', None, 'stdcall', 'wininet.FreeUrlCacheSpaceW', (('int', None), ('int', None), ('int', None)) ),
    'wininet.ftpcommanda':( 'int', None, 'stdcall', 'wininet.FtpCommandA', (('int', None), ('int', None), ('int', None), ('void *', 'ptr'), ('int', None), ('int', None)) ),
    'wininet.ftpcommandw':( 'int', None, 'stdcall', 'wininet.FtpCommandW', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wininet.ftpcreatedirectorya':( 'int', None, 'stdcall', 'wininet.FtpCreateDirectoryA', (('int', None), ('void *', 'ptr')) ),
    'wininet.ftpcreatedirectoryw':( 'int', None, 'stdcall', 'wininet.FtpCreateDirectoryW', (('int', None), ('int', None)) ),
    'wininet.ftpdeletefilea':( 'int', None, 'stdcall', 'wininet.FtpDeleteFileA', (('int', None), ('void *', 'ptr')) ),
    'wininet.ftpdeletefilew':( 'int', None, 'stdcall', 'wininet.FtpDeleteFileW', (('int', None), ('int', None)) ),
    'wininet.ftpfindfirstfilea':( 'int', None, 'stdcall', 'wininet.FtpFindFirstFileA', (('int', None), ('void *', 'ptr'), ('StringA', None), ('int', None), ('int', None)) ),
    'wininet.ftpfindfirstfilew':( 'int', None, 'stdcall', 'wininet.FtpFindFirstFileW', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wininet.ftpgetcurrentdirectorya':( 'int', None, 'stdcall', 'wininet.FtpGetCurrentDirectoryA', (('int', None), ('void *', 'ptr'), ('void *', 'ptr')) ),
    'wininet.ftpgetcurrentdirectoryw':( 'int', None, 'stdcall', 'wininet.FtpGetCurrentDirectoryW', (('int', None), ('int', None), ('int', None)) ),
    'wininet.ftpgetfilea':( 'int', None, 'stdcall', 'wininet.FtpGetFileA', ( ('int', None), ('int', None), ('int', None), ('int', None), ('DWORD', None), ('int', None)) ),
    'wininet.ftpgetfileex':( 'int', None, 'stdcall', 'wininet.FtpGetFileEx', ( ('int', None), ('StringW', None), ('int', None), ('int', None), ('DWORD', None), ('int', None)) ),
    'wininet.ftpgetfilesize':( 'int', None, 'stdcall', 'wininet.FtpGetFileSize', (('int', None), ('int', None)) ),
    'wininet.ftpgetfilew':( 'int', None, 'stdcall', 'wininet.FtpGetFileW', ( ('int', None), ('StringW', None), ('int', None), ('int', None), ('DWORD', None), ('int', None)) ),
    'wininet.ftpopenfilea':( 'int', None, 'stdcall', 'wininet.FtpOpenFileA', (('int', None), ('void *', 'ptr'), ('int', None), ('int', None), ('int', None)) ),
    'wininet.ftpopenfilew':( 'int', None, 'stdcall', 'wininet.FtpOpenFileW', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wininet.ftpputfilea':( 'int', None, 'stdcall', 'wininet.FtpPutFileA', ( ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wininet.ftpputfileex':( 'int', None, 'stdcall', 'wininet.FtpPutFileEx', ( ('StringW', None), ('int', None), ('int', None), ('int', None)) ),
    'wininet.ftpputfilew':( 'int', None, 'stdcall', 'wininet.FtpPutFileW', ( ('StringW', None), ('int', None), ('int', None), ('int', None)) ),
    'wininet.ftpremovedirectorya':( 'int', None, 'stdcall', 'wininet.FtpRemoveDirectoryA', (('int', None), ('void *', 'ptr')) ),
    'wininet.ftpremovedirectoryw':( 'int', None, 'stdcall', 'wininet.FtpRemoveDirectoryW', (('int', None), ('int', None)) ),
    'wininet.ftprenamefilea':( 'int', None, 'stdcall', 'wininet.FtpRenameFileA', (('int', None), ('void *', 'ptr'), ('void *', 'ptr')) ),
    'wininet.ftprenamefilew':( 'int', None, 'stdcall', 'wininet.FtpRenameFileW', (('int', None), ('int', None), ('int', None)) ),
    'wininet.ftpsetcurrentdirectorya':( 'int', None, 'stdcall', 'wininet.FtpSetCurrentDirectoryA', (('int', None), ('void *', 'ptr')) ),
    'wininet.ftpsetcurrentdirectoryw':( 'int', None, 'stdcall', 'wininet.FtpSetCurrentDirectoryW', (('int', None), ('int', None)) ),
    'wininet.geturlcacheconfiginfoa':( 'int', None, 'stdcall', 'wininet.GetUrlCacheConfigInfoA', (('void *', 'ptr'), ('void *', 'ptr'), ('int', None)) ),
    'wininet.geturlcacheconfiginfow':( 'int', None, 'stdcall', 'wininet.GetUrlCacheConfigInfoW', (('int', None), ('int', None), ('int', None)) ),
    'wininet.geturlcacheentryinfoa':( 'int', None, 'stdcall', 'wininet.GetUrlCacheEntryInfoA', (('int', None), ('void *', 'ptr'), ('void *', 'ptr')) ),
    'wininet.geturlcacheentryinfoexa':( 'int', None, 'stdcall', 'wininet.GetUrlCacheEntryInfoExA', (('int', None), ('void *', 'ptr'), ('void *', 'ptr'), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wininet.geturlcacheentryinfoexw':( 'int', None, 'stdcall', 'wininet.GetUrlCacheEntryInfoExW', (('void *', 'ptr'), ('void *', 'ptr'), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wininet.geturlcacheentryinfow':( 'int', None, 'stdcall', 'wininet.GetUrlCacheEntryInfoW', (('void *', 'ptr'), ('void *', 'ptr'), ('int', None)) ),
    'wininet.geturlcachegroupattributea':( 'int', None, 'stdcall', 'wininet.GetUrlCacheGroupAttributeA', (('int', None), ('int', None), ('int', None), ('int', None), ('void *', 'ptr'), ('int', None), ('int', None)) ),
    'wininet.geturlcachegroupattributew':( 'int', None, 'stdcall', 'wininet.GetUrlCacheGroupAttributeW', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wininet.geturlcacheheaderdata':( 'int', None, 'stdcall', 'wininet.GetUrlCacheHeaderData', (('int', None), ('void *', 'ptr')) ),
    'wininet.gophercreatelocatora':( 'int', None, 'stdcall', 'wininet.GopherCreateLocatorA', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wininet.gophercreatelocatorw':( 'int', None, 'stdcall', 'wininet.GopherCreateLocatorW', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wininet.gopherfindfirstfilea':( 'int', None, 'stdcall', 'wininet.GopherFindFirstFileA', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wininet.gopherfindfirstfilew':( 'int', None, 'stdcall', 'wininet.GopherFindFirstFileW', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wininet.gophergetattributea':( 'int', None, 'stdcall', 'wininet.GopherGetAttributeA', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wininet.gophergetattributew':( 'int', None, 'stdcall', 'wininet.GopherGetAttributeW', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wininet.gophergetlocatortypea':( 'int', None, 'stdcall', 'wininet.GopherGetLocatorTypeA', (('int', None), ('int', None)) ),
    'wininet.gophergetlocatortypew':( 'int', None, 'stdcall', 'wininet.GopherGetLocatorTypeW', (('int', None), ('int', None)) ),
    'wininet.gopheropenfilea':( 'int', None, 'stdcall', 'wininet.GopherOpenFileA', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wininet.gopheropenfilew':( 'int', None, 'stdcall', 'wininet.GopherOpenFileW', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wininet.httpaddrequestheadersa':( 'int', None, 'stdcall', 'wininet.HttpAddRequestHeadersA', (('int', None), ('void *', 'obj'), ('int', None), ('int', None)) ),
    'wininet.httpaddrequestheadersw':( 'int', None, 'stdcall', 'wininet.HttpAddRequestHeadersW', (('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wininet.httpcheckdavcompliance':( 'int', None, 'stdcall', 'wininet.HttpCheckDavCompliance', (('void *', 'ptr'), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wininet.httpendrequesta':( 'int', None, 'stdcall', 'wininet.HttpEndRequestA', (('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wininet.httpendrequestw':( 'int', None, 'stdcall', 'wininet.HttpEndRequestW', (('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wininet.httpopenrequesta':( 'int', None, 'stdcall', 'wininet.HttpOpenRequestA', (('int', None), ('int', None), ('void *', 'ptr'), ('void *', 'ptr'), ('void *', 'ptr'), ('int', None), ('int', None), ('int', None)) ),
    'wininet.httpopenrequestw':( 'int', None, 'stdcall', 'wininet.HttpOpenRequestW', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wininet.httpqueryinfoa':( 'int', None, 'stdcall', 'wininet.HttpQueryInfoA', (('int', None), ('int', None), ('void *', 'obj'), ('void *', 'ptr'), ('void *', 'ptr')) ),
    'wininet.httpqueryinfow':( 'int', None, 'stdcall', 'wininet.HttpQueryInfoW', (('int', None), ('int', None), ('void *', 'obj'), ('int', None), ('void *', 'ptr')) ),
    'wininet.httpsendrequesta':( 'int', None, 'stdcall', 'wininet.HttpSendRequestA', (('int', None), ('void *', 'obj'), ('int', None), ('int', None), ('int', None)) ),
    'wininet.httpsendrequestexa':( 'int', None, 'stdcall', 'wininet.HttpSendRequestExA', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wininet.httpsendrequestexw':( 'int', None, 'stdcall', 'wininet.HttpSendRequestExW', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wininet.httpsendrequestw':( 'int', None, 'stdcall', 'wininet.HttpSendRequestW', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wininet.incrementurlcacheheaderdata':( 'int', None, 'stdcall', 'wininet.IncrementUrlCacheHeaderData', (('int', None), ('int', None)) ),
    'wininet.internetalgidtostringa':( 'int', None, 'stdcall', 'wininet.InternetAlgIdToStringA', (('int', None), ('void *', 'ptr'), ('int', None), ('int', None)) ),
    'wininet.internetalgidtostringw':( 'int', None, 'stdcall', 'wininet.InternetAlgIdToStringW', (('int', None), ('void *', 'ptr'), ('int', None), ('int', None)) ),
    'wininet.internetattemptconnect':( 'int', None, 'stdcall', 'wininet.InternetAttemptConnect', (('int', None),) ),
    'wininet.internetautodial':( 'int', None, 'stdcall', 'wininet.InternetAutodial', (('int', None), ('int', None)) ),
    'wininet.internetautodialcallback':( 'int', None, 'stdcall', 'wininet.InternetAutodialCallback', (('int', None), ('void *', 'ptr')) ),
    'wininet.internetautodialhangup':( 'int', None, 'stdcall', 'wininet.InternetAutodialHangup', (('int', None),) ),
    'wininet.internetcanonicalizeurla':( 'int', None, 'stdcall', 'wininet.InternetCanonicalizeUrlA', (('void *', 'ptr'), ('int', None), ('void *', 'ptr'), ('int', None)) ),
    'wininet.internetcanonicalizeurlw':( 'int', None, 'stdcall', 'wininet.InternetCanonicalizeUrlW', (('int', None), ('int', None), ('void *', 'ptr'), ('int', None)) ),
    'wininet.internetcheckconnectiona':( 'int', None, 'stdcall', 'wininet.InternetCheckConnectionA', (('void *', 'ptr'), ('int', None), ('int', None)) ),
    'wininet.internetcheckconnectionw':( 'int', None, 'msfastcall', 'wininet.InternetCheckConnectionW', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wininet.internetclearallpersitecookiedecisions':( 'int', None, 'cdecl', 'wininet.InternetClearAllPerSiteCookieDecisions', () ),
    'wininet.internetclosehandle':( 'int', None, 'stdcall', 'wininet.InternetCloseHandle', (('int', None),) ),
    'wininet.internetcombineurla':( 'int', None, 'stdcall', 'wininet.InternetCombineUrlA', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wininet.internetcombineurlw':( 'int', None, 'stdcall', 'wininet.InternetCombineUrlW', (('int', None), ('int', None), ('int', None), ('void *', 'ptr'), ('int', None)) ),
    'wininet.internetconfirmzonecrossing':( 'int', None, 'stdcall', 'wininet.InternetConfirmZoneCrossing', (('int', None), ('void *', 'obj'), ('int', None), ('int', None)) ),
    'wininet.internetconfirmzonecrossinga':( 'int', None, 'stdcall', 'wininet.InternetConfirmZoneCrossingA', (('int', None), ('void *', 'obj'), ('int', None), ('int', None)) ),
    'wininet.internetconfirmzonecrossingw':( 'int', None, 'stdcall', 'wininet.InternetConfirmZoneCrossingW', (('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wininet.internetconnecta':( 'int', None, 'stdcall', 'wininet.InternetConnectA', ( ('void *', 'ptr'), ('int', None), ('void *', 'ptr'), ('void *', 'ptr'), ('int', None), ('int', None), ('int', None)) ),
    'wininet.internetconnectw':( 'int', None, 'stdcall', 'wininet.InternetConnectW', ( ('void *', 'ptr'), ('int', None), ('void *', 'ptr'), ('void *', 'ptr'), ('int', None), ('int', None), ('int', None)) ),
    'wininet.internetcrackurla':( 'int', None, 'stdcall', 'wininet.InternetCrackUrlA', (('void *', 'ptr'), ('int', None), ('int', None), ('void *', 'ptr')) ),
    'wininet.internetcrackurlw':( 'int', None, 'stdcall', 'wininet.InternetCrackUrlW', (('int', None), ('int', None), ('int', None), ('void *', 'ptr')) ),
    'wininet.internetcreateurla':( 'int', None, 'stdcall', 'wininet.InternetCreateUrlA', (('void *', 'ptr'), ('int', None), ('void *', 'ptr'), ('void *', 'ptr')) ),
    'wininet.internetcreateurlw':( 'int', None, 'stdcall', 'wininet.InternetCreateUrlW', (('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wininet.internetdial':( 'int', None, 'stdcall', 'wininet.InternetDial', (('int', None), ('int', None), ('int', None), ('void *', 'ptr'), ('int', None)) ),
    'wininet.internetdiala':( 'int', None, 'stdcall', 'wininet.InternetDialA', (('int', None), ('int', None), ('int', None), ('void *', 'ptr'), ('int', None)) ),
    'wininet.internetdialw':( 'int', None, 'stdcall', 'wininet.InternetDialW', (('int', None), ('void *', 'ptr'), ('int', None), ('void *', 'ptr'), ('int', None)) ),
    'wininet.internetenumpersitecookiedecisiona':( 'int', None, 'stdcall', 'wininet.InternetEnumPerSiteCookieDecisionA', (('int', None), ('void *', 'ptr'), ('void *', 'ptr'), ('int', None)) ),
    'wininet.internetenumpersitecookiedecisionw':( 'int', None, 'stdcall', 'wininet.InternetEnumPerSiteCookieDecisionW', (('int', None), ('void *', 'ptr'), ('void *', 'ptr'), ('int', None)) ),
    'wininet.interneterrordlg':( 'int', None, 'stdcall', 'wininet.InternetErrorDlg', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wininet.internetfindnextfilea':( 'int', None, 'stdcall', 'wininet.InternetFindNextFileA', (('int', None), ('StringA', None)) ),
    'wininet.internetfindnextfilew':( 'int', None, 'stdcall', 'wininet.InternetFindNextFileW', (('int', None), ('int', None)) ),
    'wininet.internetfortezzacommand':( 'int', None, 'stdcall', 'wininet.InternetFortezzaCommand', (('int', None), ('int', None), ('int', None)) ),
    'wininet.internetgetcertbyurl':( 'int', None, 'stdcall', 'wininet.InternetGetCertByURL', (('int', None), ('int', None), ('int', None)) ),
    'wininet.internetgetcertbyurla':( 'int', None, 'stdcall', 'wininet.InternetGetCertByURLA', (('int', None), ('int', None), ('int', None)) ),
    'wininet.internetgetconnectedstate':( 'int', None, 'stdcall', 'wininet.InternetGetConnectedState', (('void *', 'ptr'), ('int', None)) ),
    'wininet.internetgetconnectedstateex':( 'int', None, 'stdcall', 'wininet.InternetGetConnectedStateEx', (('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wininet.internetgetconnectedstateexa':( 'int', None, 'stdcall', 'wininet.InternetGetConnectedStateExA', (('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wininet.internetgetconnectedstateexw':( 'int', None, 'stdcall', 'wininet.InternetGetConnectedStateExW', (('int', None), ('void *', 'ptr'), ('int', None), ('int', None)) ),
    'wininet.internetgetcookiea':( 'int', None, 'msfastcall', 'wininet.InternetGetCookieA', (('int', None), ('int', None), ('void *', 'ptr'), ('int', None), ('int', None), ('void *', 'ptr')) ),
    'wininet.internetgetcookieexa':( 'int', None, 'msfastcall', 'wininet.InternetGetCookieExA', (('void *', 'ptr'), ('int', None), ('void *', 'ptr'), ('void *', 'ptr'), ('int', None), ('void *', 'ptr'), ('int', None), ('int', None)) ),
    'wininet.internetgetcookieexw':( 'int', None, 'msfastcall', 'wininet.InternetGetCookieExW', (('int', None), ('int', None), ('void *', 'ptr'), ('int', None), ('int', None), ('void *', 'ptr'), ('int', None), ('int', None)) ),
    'wininet.internetgetcookiew':( 'int', None, 'msfastcall', 'wininet.InternetGetCookieW', (('int', None), ('int', None), ('void *', 'ptr'), ('int', None), ('int', None), ('void *', 'ptr')) ),
    'wininet.internetgetlastresponseinfoa':( 'int', None, 'stdcall', 'wininet.InternetGetLastResponseInfoA', (('void *', 'ptr'), ('int', None), ('void *', 'ptr')) ),
    'wininet.internetgetlastresponseinfow':( 'int', None, 'stdcall', 'wininet.InternetGetLastResponseInfoW', (('int', None), ('int', None), ('int', None)) ),
    'wininet.internetgetpersitecookiedecisiona':( 'int', None, 'stdcall', 'wininet.InternetGetPerSiteCookieDecisionA', (('void *', 'ptr'), ('int', None)) ),
    'wininet.internetgetpersitecookiedecisionw':( 'int', None, 'msfastcall', 'wininet.InternetGetPerSiteCookieDecisionW', (('int', None), ('int', None), ('int', None), ('void *', 'ptr')) ),
    'wininet.internetgetsecurityinfobyurl':( 'int', None, 'stdcall', 'wininet.InternetGetSecurityInfoByURL', ( ('void *', 'ptr'), ('int', None), ('int', None)) ),
    'wininet.internetgetsecurityinfobyurla':( 'int', None, 'stdcall', 'wininet.InternetGetSecurityInfoByURLA', ( ('void *', 'ptr'), ('int', None), ('int', None)) ),
    'wininet.internetgetsecurityinfobyurlw':( 'int', None, 'stdcall', 'wininet.InternetGetSecurityInfoByURLW', ( ('void *', 'ptr'), ('int', None), ('int', None)) ),
    'wininet.internetgoonline':( 'int', None, 'msfastcall', 'wininet.InternetGoOnline', (('int', None), ('int', None), ('void *', 'ptr'), ('int', None), ('int', None)) ),
    'wininet.internetgoonlinea':( 'int', None, 'msfastcall', 'wininet.InternetGoOnlineA', (('int', None), ('int', None), ('void *', 'ptr'), ('int', None), ('int', None)) ),
    'wininet.internetgoonlinew':( 'int', None, 'msfastcall', 'wininet.InternetGoOnlineW', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wininet.internethangup':( 'int', None, 'stdcall', 'wininet.InternetHangUp', (('int', None), ('int', None)) ),
    'wininet.internetinitializeautoproxydll':( 'int', None, 'stdcall', 'wininet.InternetInitializeAutoProxyDll', ( ('int', None),) ),
    'wininet.internetlockrequestfile':( 'int', None, 'stdcall', 'wininet.InternetLockRequestFile', ( ('void *', 'obj'), ('void *', 'ptr')) ),
    'wininet.internetopena':( 'int', None, 'stdcall', 'wininet.InternetOpenA', ( ('void *', 'ptr'), ('int', None), ('void *', 'ptr'), ('void *', 'ptr'), ('int', None)) ),
    'wininet.internetopenurla':( 'int', None, 'stdcall', 'wininet.InternetOpenUrlA', (('int', None), ('void *', 'ptr'), ('void *', 'ptr'), ('int', None), ('int', None), ('int', None)) ),
    'wininet.internetopenurlw':( 'int', None, 'stdcall', 'wininet.InternetOpenUrlW', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wininet.internetopenw':( 'int', None, 'stdcall', 'wininet.InternetOpenW', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wininet.internetquerydataavailable':( 'int', None, 'stdcall', 'wininet.InternetQueryDataAvailable', (('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wininet.internetqueryfortezzastatus':( 'int', None, 'stdcall', 'wininet.InternetQueryFortezzaStatus', (('int', None), ('int', None)) ),
    'wininet.internetqueryoptiona':( 'int', None, 'stdcall', 'wininet.InternetQueryOptionA', ( ('void *', 'obj'), ('int', None), ('void *', 'obj'), ('void *', 'ptr')) ),
    'wininet.internetqueryoptionw':( 'int', None, 'msfastcall', 'wininet.InternetQueryOptionW', (('int', None), ('int', None), ('void *', 'obj'), ('int', None), ('void *', 'obj'), ('void *', 'ptr')) ),
    'wininet.internetreadfile':( 'int', None, 'stdcall', 'wininet.InternetReadFile', (('int', None), ('StringA', None), ('DWORD', None), ('void *', 'ptr')) ),
    'wininet.internetreadfileexa':( 'int', None, 'stdcall', 'wininet.InternetReadFileExA', (('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wininet.internetreadfileexw':( 'int', None, 'stdcall', 'wininet.InternetReadFileExW', (('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wininet.internetsecurityprotocoltostringa':( 'int', None, 'stdcall', 'wininet.InternetSecurityProtocolToStringA', (('int', None), ('void *', 'ptr'), ('int', None), ('int', None)) ),
    'wininet.internetsecurityprotocoltostringw':( 'int', None, 'stdcall', 'wininet.InternetSecurityProtocolToStringW', (('int', None), ('void *', 'ptr'), ('int', None), ('int', None)) ),
    'wininet.internetsetcookiea':( 'int', None, 'stdcall', 'wininet.InternetSetCookieA', (('void *', 'ptr'), ('int', None), ('void *', 'ptr')) ),
    'wininet.internetsetcookieexa':( 'int', None, 'stdcall', 'wininet.InternetSetCookieExA', (('void *', 'ptr'), ('int', None), ('void *', 'ptr'), ('int', None), ('int', None)) ),
    'wininet.internetsetcookieexw':( 'int', None, 'stdcall', 'wininet.InternetSetCookieExW', (('void *', 'ptr'), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wininet.internetsetcookiew':( 'int', None, 'stdcall', 'wininet.InternetSetCookieW', (('void *', 'ptr'), ('int', None), ('int', None)) ),
    'wininet.internetsetdialstate':( 'int', None, 'stdcall', 'wininet.InternetSetDialState', (('int', None), ('int', None), ('int', None)) ),
    'wininet.internetsetdialstatea':( 'int', None, 'stdcall', 'wininet.InternetSetDialStateA', (('int', None), ('int', None), ('int', None)) ),
    'wininet.internetsetdialstatew':( 'int', None, 'stdcall', 'wininet.InternetSetDialStateW', (('int', None), ('int', None), ('int', None)) ),
    'wininet.internetsetfilepointer':( 'int', None, 'stdcall', 'wininet.InternetSetFilePointer', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wininet.internetsetoptiona':( 'int', None, 'stdcall', 'wininet.InternetSetOptionA', ( ('void *', 'obj'), ('int', None), ('void *', 'ptr'), ('int', None)) ),
    'wininet.internetsetoptionexa':( 'int', None, 'stdcall', 'wininet.InternetSetOptionExA', ( ('void *', 'obj'), ('int', None), ('void *', 'ptr'), ('int', None), ('int', None)) ),
    'wininet.internetsetoptionexw':( 'int', None, 'msfastcall', 'wininet.InternetSetOptionExW', (('int', None), ('int', None), ('void *', 'obj'), ('int', None), ('void *', 'ptr'), ('int', None), ('int', None)) ),
    'wininet.internetsetoptionw':( 'int', None, 'msfastcall', 'wininet.InternetSetOptionW', (('int', None), ('int', None), ('void *', 'obj'), ('int', None), ('void *', 'ptr'), ('int', None)) ),
    'wininet.internetsetpersitecookiedecisiona':( 'int', None, 'stdcall', 'wininet.InternetSetPerSiteCookieDecisionA', (('int', None), ('int', None)) ),
    'wininet.internetsetpersitecookiedecisionw':( 'int', None, 'msfastcall', 'wininet.InternetSetPerSiteCookieDecisionW', (('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wininet.internetsetstatuscallback':( 'int', None, 'stdcall', 'wininet.InternetSetStatusCallback', ( ('int', None), ('int', None)) ),
    'wininet.internetsetstatuscallbacka':( 'int', None, 'stdcall', 'wininet.InternetSetStatusCallbackA', ( ('int', None), ('int', None)) ),
    'wininet.internetsetstatuscallbackw':( 'int', None, 'stdcall', 'wininet.InternetSetStatusCallbackW', ( ('int', None), ('int', None)) ),
    'wininet.internetshowsecurityinfobyurl':( 'int', None, 'stdcall', 'wininet.InternetShowSecurityInfoByURL', (('void *', 'ptr'), ('int', None)) ),
    'wininet.internetshowsecurityinfobyurla':( 'int', None, 'stdcall', 'wininet.InternetShowSecurityInfoByURLA', (('void *', 'ptr'), ('int', None)) ),
    'wininet.internetshowsecurityinfobyurlw':( 'int', None, 'stdcall', 'wininet.InternetShowSecurityInfoByURLW', (('void *', 'ptr'), ('int', None)) ),
    'wininet.internettimefromsystemtime':( 'int', None, 'stdcall', 'wininet.InternetTimeFromSystemTime', (('void *', 'ptr'), ('int', None), ('void *', 'ptr'), ('int', None)) ),
    'wininet.internettimefromsystemtimea':( 'int', None, 'stdcall', 'wininet.InternetTimeFromSystemTimeA', (('void *', 'ptr'), ('int', None), ('void *', 'ptr'), ('int', None)) ),
    'wininet.internettimefromsystemtimew':( 'int', None, 'stdcall', 'wininet.InternetTimeFromSystemTimeW', (('void *', 'ptr'), ('int', None), ('int', None), ('DWORD', None)) ),
    'wininet.internettimetosystemtime':( 'int', None, 'stdcall', 'wininet.InternetTimeToSystemTime', (('int', None), ('void *', 'ptr'), ('int', None)) ),
    'wininet.internettimetosystemtimea':( 'int', None, 'stdcall', 'wininet.InternetTimeToSystemTimeA', (('int', None), ('void *', 'ptr'), ('int', None)) ),
    'wininet.internettimetosystemtimew':( 'int', None, 'stdcall', 'wininet.InternetTimeToSystemTimeW', (('int', None), ('int', None), ('int', None)) ),
    'wininet.internetunlockrequestfile':( 'int', None, 'stdcall', 'wininet.InternetUnlockRequestFile', ( ('void *', 'ptr'),) ),
    'wininet.internetwritefile':( 'int', None, 'stdcall', 'wininet.InternetWriteFile', ( ('int', None), ('void *', 'ptr'), ('DWORD', None), ('DWORD_P', None)) ),
    'wininet.internetwritefileexa':( 'int', None, 'stdcall', 'wininet.InternetWriteFileExA', (('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wininet.internetwritefileexw':( 'int', None, 'stdcall', 'wininet.InternetWriteFileExW', (('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wininet.ishostinproxybypasslist':( 'int', None, 'stdcall', 'wininet.IsHostInProxyBypassList', ( ('int', None), ('void *', 'ptr'), ('int', None)) ),
    'wininet.isurlcacheentryexpireda':( 'int', None, 'stdcall', 'wininet.IsUrlCacheEntryExpiredA', (('int', None), ('int', None), ('int', None)) ),
    'wininet.isurlcacheentryexpiredw':( 'int', None, 'stdcall', 'wininet.IsUrlCacheEntryExpiredW', (('void *', 'ptr'), ('int', None), ('int', None)) ),
    'wininet.loadurlcachecontent':( 'int', None, 'cdecl', 'wininet.LoadUrlCacheContent', () ),
    'wininet.parsex509encodedcertificateforlistboxentry':( 'int', None, 'stdcall', 'wininet.ParseX509EncodedCertificateForListBoxEntry', (('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wininet.privacygetzonepreferencew':( 'int', None, 'stdcall', 'wininet.PrivacyGetZonePreferenceW', (('int', None), ('int', None), ('int', None), ('int', None), ('void *', 'ptr')) ),
    'wininet.privacysetzonepreferencew':( 'int', None, 'stdcall', 'wininet.PrivacySetZonePreferenceW', (('int', None), ('int', None), ('int', None), ('void *', 'ptr')) ),
    'wininet.readurlcacheentrystream':( 'int', None, 'stdcall', 'wininet.ReadUrlCacheEntryStream', (('int', None), ('int', None), ('StringA', None), ('DWORD_P', None), ('int', None)) ),
    'wininet.readurlcacheentrystreamex':( 'int', None, 'stdcall', 'wininet.ReadUrlCacheEntryStreamEx', (('int', None), ('int', None), ('int', None), ('StringA', None), ('DWORD_P', None)) ),
    'wininet.registerurlcachenotification':( 'int', None, 'stdcall', 'wininet.RegisterUrlCacheNotification', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wininet.resumesuspendeddownload':( 'int', None, 'stdcall', 'wininet.ResumeSuspendedDownload', (('int', None), ('int', None)) ),
    'wininet.retrieveurlcacheentryfilea':( 'int', None, 'stdcall', 'wininet.RetrieveUrlCacheEntryFileA', (('int', None), ('int', None), ('void *', 'ptr'), ('int', None)) ),
    'wininet.retrieveurlcacheentryfilew':( 'int', None, 'stdcall', 'wininet.RetrieveUrlCacheEntryFileW', (('void *', 'ptr'), ('void *', 'ptr'), ('int', None), ('int', None)) ),
    'wininet.retrieveurlcacheentrystreama':( 'int', None, 'stdcall', 'wininet.RetrieveUrlCacheEntryStreamA', (('void *', 'ptr'), ('int', None), ('void *', 'ptr'), ('int', None), ('int', None)) ),
    'wininet.retrieveurlcacheentrystreamw':( 'int', None, 'stdcall', 'wininet.RetrieveUrlCacheEntryStreamW', (('void *', 'ptr'), ('void *', 'ptr'), ('int', None), ('int', None), ('int', None)) ),
    'wininet.runonceurlcache':( 'int', None, 'stdcall', 'wininet.RunOnceUrlCache', (('int', None), ('int', None), ('void *', 'ptr'), ('int', None)) ),
    'wininet.seturlcacheconfiginfoa':( 'int', None, 'stdcall', 'wininet.SetUrlCacheConfigInfoA', (('void *', 'ptr'), ('int', None)) ),
    'wininet.seturlcacheconfiginfow':( 'int', None, 'stdcall', 'wininet.SetUrlCacheConfigInfoW', (('int', None), ('int', None)) ),
    'wininet.seturlcacheentrygroup':( 'int', None, 'stdcall', 'wininet.SetUrlCacheEntryGroup', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wininet.seturlcacheentrygroupa':( 'int', None, 'stdcall', 'wininet.SetUrlCacheEntryGroupA', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wininet.seturlcacheentrygroupw':( 'int', None, 'stdcall', 'wininet.SetUrlCacheEntryGroupW', (('void *', 'ptr'), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wininet.seturlcacheentryinfoa':( 'int', None, 'stdcall', 'wininet.SetUrlCacheEntryInfoA', (('void *', 'ptr'), ('void *', 'ptr'), ('int', None)) ),
    'wininet.seturlcacheentryinfow':( 'int', None, 'stdcall', 'wininet.SetUrlCacheEntryInfoW', (('void *', 'ptr'), ('int', None), ('int', None)) ),
    'wininet.seturlcachegroupattributea':( 'int', None, 'stdcall', 'wininet.SetUrlCacheGroupAttributeA', (('int', None), ('int', None), ('int', None), ('int', None), ('void *', 'ptr'), ('int', None)) ),
    'wininet.seturlcachegroupattributew':( 'int', None, 'stdcall', 'wininet.SetUrlCacheGroupAttributeW', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'wininet.seturlcacheheaderdata':( 'int', None, 'stdcall', 'wininet.SetUrlCacheHeaderData', (('int', None), ('int', None)) ),
    'wininet.showcertificate':( 'int', None, 'stdcall', 'wininet.ShowCertificate', (('int', None), ('int', None)) ),
    'wininet.showclientauthcerts':( 'int', None, 'stdcall', 'wininet.ShowClientAuthCerts', (('int', None),) ),
    'wininet.showsecurityinfo':( 'int', None, 'stdcall', 'wininet.ShowSecurityInfo', (('int', None), ('void *', 'ptr')) ),
    'wininet.showx509encodedcertificate':( 'int', None, 'stdcall', 'wininet.ShowX509EncodedCertificate', (('int', None), ('int', None), ('int', None)) ),
    'wininet.unlockurlcacheentryfile':( 'int', None, 'stdcall', 'wininet.UnlockUrlCacheEntryFile', (('int', None), ('int', None)) ),
    'wininet.unlockurlcacheentryfilea':( 'int', None, 'stdcall', 'wininet.UnlockUrlCacheEntryFileA', (('int', None), ('int', None)) ),
    'wininet.unlockurlcacheentryfilew':( 'int', None, 'stdcall', 'wininet.UnlockUrlCacheEntryFileW', (('void *', 'ptr'), ('int', None)) ),
    'wininet.unlockurlcacheentrystream':( 'int', None, 'stdcall', 'wininet.UnlockUrlCacheEntryStream', (('int', None), ('int', None)) ),
    'wininet.updateurlcachecontentpath':( 'int', None, 'stdcall', 'wininet.UpdateUrlCacheContentPath', (('StringA', None),) ),
    'wininet.urlzonesdetach':( 'int', None, 'cdecl', 'wininet.UrlZonesDetach', () ),
    'wininet._getfileextensionfromurl':( 'int', None, 'stdcall', 'wininet._GetFileExtensionFromUrl', (('int', None), ('int', None), ('void *', 'ptr'), ('int', None)) ),
    }
